# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the mergeuploads task support on the worker."""

import hashlib
from collections.abc import Iterable
from pathlib import Path
from typing import override
from unittest import mock
from unittest.mock import call

from debian import deb822

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts import Upload
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianUpload,
    WorkRequestResults,
)
from debusine.artifacts.playground import ArtifactPlayground
from debusine.client.models import LookupSingleResponseArtifact
from debusine.tasks import MergeUploads, TaskConfigError
from debusine.tasks.mergeuploads import MergeUploadsError
from debusine.tasks.models import BaseTaskInputArtifact, MergeUploadsDynamicData
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_artifact_response,
    create_remote_artifact,
)
from debusine.utils import read_changes


class MergeUploadsTests(ExternalTaskHelperMixin[MergeUploads], TestCase):
    """Test the MergeUploads class."""

    SAMPLE_TASK_DATA = {"input": {"uploads": [421, 666]}}

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifacts(
            "uploads",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "amd64",
                        "Files": [{"name": "hello_1.0_amd64.deb"}],
                        "Source": "hello",
                    },
                ),
            ),
            BaseTaskInputArtifact(
                artifact_id=666,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "s390x",
                        "Files": [{"name": "hello_1.0_s390x.deb"}],
                        "Source": "hello",
                    },
                ),
            ),
        )

    @staticmethod
    def _make_files_field(paths: Iterable[Path]) -> list[dict[str, str]]:
        """Make a valid ``Files`` field."""
        return [
            {
                "md5sum": hashlib.md5(path.read_bytes()).hexdigest(),
                "name": path.name,
                "priority": "optional",
                "section": "unknown",
                "size": str(path.stat().st_size),
            }
            for path in paths
        ]

    @staticmethod
    def _make_checksums_field(
        hash_name: str, paths: Iterable[Path]
    ) -> list[dict[str, str]]:
        """Make a valid ``Checksums-*`` field."""
        return [
            {
                "name": path.name,
                hash_name: hashlib.new(
                    hash_name, path.read_bytes()
                ).hexdigest(),
                "size": str(path.stat().st_size),
            }
            for path in paths
        ]

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [
                wtags.TASK_PREFIX + "worker:mergeuploads:version:1",
            ],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.SOURCE_PACKAGE_PREFIX + "hello"},
        )

    def test_compute_scheduler_tags_multiple_packages(self) -> None:
        self.resolver.set_field_artifacts(
            "uploads",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "amd64",
                        "Files": [{"name": "hello_1.0_amd64.deb"}],
                        "Source": "hello",
                    },
                ),
            ),
            BaseTaskInputArtifact(
                artifact_id=666,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "s390x",
                        "Files": [{"name": "bye_1.0-1_s390x.deb"}],
                        "Source": "bye",
                    },
                ),
            ),
        )

        self.assertComputesSchedulerTags(
            provide_system={
                ttags.SOURCE_PACKAGE_PREFIX + "hello",
                ttags.SOURCE_PACKAGE_PREFIX + "bye",
            }
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            MergeUploadsDynamicData(
                input_uploads_ids=[421, 666],
                subject="hello",
                parameter_summary="hello (amd64, s390x)",
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_compute_dynamic_data_multiple_packages(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.resolver.set_field_artifacts(
            "uploads",
            BaseTaskInputArtifact(
                artifact_id=421,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "amd64",
                        "Files": [{"name": "hello_1.0_amd64.deb"}],
                        "Source": "hello",
                    },
                ),
            ),
            BaseTaskInputArtifact(
                artifact_id=666,
                category=ArtifactCategory.UPLOAD,
                data=DebianUpload(
                    type="dpkg",
                    changes_fields={
                        "Architecture": "s390x",
                        "Files": [{"name": "bye_1.0-1_s390x.deb"}],
                        "Source": "bye",
                    },
                ),
            ),
        )

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            MergeUploadsDynamicData(
                input_uploads_ids=[421, 666],
                subject=None,
                parameter_summary=None,
            ),
        )

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            MergeUploads.inputs["uploads"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.UPLOAD,),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = MergeUploadsDynamicData(
            input_uploads_ids=[2, 3],
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [2, 3])

    def test_configure_fails_with_missing_required_data(
        self,
    ) -> None:
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"input": {}})

    def test_fetch_input(self) -> None:
        """Test fetch_input: call fetch_artifact(artifact_id, directory)."""
        directory = self.create_temporary_directory()
        artifact1 = self.fake_debian_upload_artifact()
        artifact2 = self.fake_debian_upload_artifact()
        artifact2.id += 1
        self.configure_task()
        self.task.work_request_id = 5
        self.task.dynamic_data = MergeUploadsDynamicData(
            input_uploads_ids=[artifact1.id, artifact2.id]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.artifact_get.return_value = artifact1

        with mock.patch.object(
            self.task, "fetch_artifact", autospec=True, return_value=True
        ) as fetch_artifact_mocked:
            result = self.task.fetch_input(directory)

        self.assertTrue(result)
        fetch_artifact_mocked.assert_has_calls(
            [
                mock.call(artifact1.id, directory / f"upload-{artifact1.id}"),
                mock.call(artifact2.id, directory / f"upload-{artifact2.id}"),
            ]
        )

    def test_fetch_input_wrong_category(self) -> None:
        """Test fetch_input when input isn't a package upload."""
        directory = self.create_temporary_directory()
        artifact1 = create_artifact_response(id=12)
        artifact2 = self.fake_debian_upload_artifact()
        self.configure_task()
        self.task.work_request_id = 5
        self.task.dynamic_data = MergeUploadsDynamicData(
            input_uploads_ids=[artifact1.id, artifact2.id]
        )
        debusine_mock = self.mock_debusine()
        debusine_mock.artifact_get.return_value = artifact1

        with mock.patch.object(
            self.task, "fetch_artifact", autospec=True
        ) as fetch_artifact_mocked:
            result = self.task.fetch_input(directory)

        self.assertFalse(result)

        log_file_contents = (
            self.task.log_directory / "fetch_input.log"
        ).read_text()
        self.assertEqual(
            log_file_contents,
            (
                "input.uploads points to a Testing, not the "
                "expected debian:upload.\n"
            ),
        )

        fetch_artifact_mocked.assert_not_called()

    def test_configure_for_execution_groups_changes_by_upload(self) -> None:
        directory = self.create_temporary_directory()
        changes_source = directory / "upload-9" / "foo_1.0_source.changes"
        changes_amd64 = directory / "upload-10" / "foo_1.0_amd64.changes"
        changes_i386 = directory / "upload-11" / "foo_1.0_i386.changes"
        for changes in (changes_source, changes_amd64, changes_i386):
            changes.parent.mkdir()
            changes.touch()

        self.task.configure_for_execution(directory)

        self.assertEqual(
            self.task._changes_paths,
            [changes_source, changes_amd64, changes_i386],
        )

    def test_merge_changes_unknown_format(self) -> None:
        """Only ``Format: 1.8`` is supported."""
        execute_directory = self.create_temporary_directory()
        with self.assertRaisesRegex(
            MergeUploadsError, "Unknown .changes format: 1.7"
        ):
            self.task.merge_changes(
                execute_directory, [deb822.Changes({"Format": "1.7"})]
            )

    def test_merge_changes_mismatched_simple_fields(self) -> None:
        """All ``Format``, ``Source``, and ``Version`` fields must match."""
        for field in ("Format", "Source", "Version"):
            with self.subTest(field=field):
                execute_directory = self.create_temporary_directory()
                all_changes = [
                    deb822.Changes(
                        {"Format": "1.8", "Source": "hello", "Version": "1.0"}
                    )
                    for _ in range(2)
                ]
                all_changes[1][field] = "mismatch"
                with self.assertRaisesRegex(
                    MergeUploadsError,
                    rf"{field} fields do not match: "
                    rf"\[{all_changes[0][field]!r}, 'mismatch'\]",
                ):
                    self.task.merge_changes(execute_directory, all_changes)

    def test_merge_changes_mismatched_descriptions(self) -> None:
        """The descriptions for each binary package must match."""
        execute_directory = self.create_temporary_directory()
        with self.assertRaisesRegex(
            MergeUploadsError,
            "Descriptions for mismatch do not match: 'left' != 'right'",
        ):
            self.task.merge_changes(
                execute_directory,
                [
                    deb822.Changes(
                        {
                            "Format": "1.8",
                            "Source": "hello",
                            "Version": "1.0",
                            "Description": (
                                "\n"
                                " unrecognized junk\n"
                                " match - matching description\n"
                                " unique-left - only in left\n"
                                " mismatch - left"
                            ),
                        }
                    ),
                    deb822.Changes(
                        {
                            "Format": "1.8",
                            "Source": "hello",
                            "Version": "1.0",
                            "Description": (
                                "\n"
                                " unrecognized junk\n"
                                " match - matching description\n"
                                " unique-right - only in right\n"
                                " mismatch - right"
                            ),
                        }
                    ),
                ],
            )

    def test_merge_changes_mismatched_files(self) -> None:
        """The entries for a given file in ``Files`` must match."""
        execute_directory = self.create_temporary_directory()
        with self.assertRaisesRegex(
            MergeUploadsError,
            r"Entries in Files for mismatch_1\.0_all.deb do not match: "
            r"\{'md5sum': '0002', 'size': '24', 'section': 'misc', "
            r"'priority': 'optional', 'name': 'mismatch_1\.0_all.deb'\} != "
            r"\{'md5sum': '0003', 'size': '32', 'section': 'misc', "
            r"'priority': 'optional', 'name': 'mismatch_1\.0_all.deb'\}",
        ):
            self.task.merge_changes(
                execute_directory,
                [
                    deb822.Changes(
                        {
                            "Format": "1.8",
                            "Source": "hello",
                            "Version": "1.0",
                            "Files": (
                                "\n"
                                " 0000 16 misc optional archdep_1.0_amd64.deb\n"
                                " 0001 20 misc optional match_1.0_all.deb\n"
                                " 0002 24 misc optional mismatch_1.0_all.deb"
                            ),
                        }
                    ),
                    deb822.Changes(
                        {
                            "Format": "1.8",
                            "Source": "hello",
                            "Version": "1.0",
                            "Files": (
                                "\n"
                                " 0000 16 misc optional archdep_1.0_amd64.deb\n"
                                " 0001 20 misc optional match_1.0_all.deb\n"
                                " 0003 32 misc optional mismatch_1.0_all.deb"
                            ),
                        }
                    ),
                ],
            )

    def test_merge_changes_mismatched_checksums(self) -> None:
        """The entries for a given file in ``Checksums-*`` must match."""
        for field, checksum in (
            ("Checksums-Sha1", "sha1"),
            ("Checksums-Sha256", "sha256"),
        ):
            with self.subTest(field=field):
                execute_directory = self.create_temporary_directory()
                with self.assertRaisesRegex(
                    MergeUploadsError,
                    rf"Entries in {field} for mismatch_1\.0_all.deb do not "
                    rf"match: "
                    rf"\{{'{checksum}': '0002', 'size': '24', "
                    rf"'name': 'mismatch_1\.0_all.deb'\}} != "
                    rf"\{{'{checksum}': '0003', 'size': '32', "
                    rf"'name': 'mismatch_1\.0_all.deb'\}}",
                ):
                    self.task.merge_changes(
                        execute_directory,
                        [
                            deb822.Changes(
                                {
                                    "Format": "1.8",
                                    "Source": "hello",
                                    "Version": "1.0",
                                    field: (
                                        "\n"
                                        " 0000 16 archdep_1.0_amd64.deb\n"
                                        " 0001 20 match_1.0_all.deb\n"
                                        " 0002 24 mismatch_1.0_all.deb"
                                    ),
                                }
                            ),
                            deb822.Changes(
                                {
                                    "Format": "1.8",
                                    "Source": "hello",
                                    "Version": "1.0",
                                    field: (
                                        "\n"
                                        " 0000 16 archdep_1.0_amd64.deb\n"
                                        " 0001 20 match_1.0_all.deb\n"
                                        " 0003 32 mismatch_1.0_all.deb"
                                    ),
                                }
                            ),
                        ],
                    )

    def test_merge_changes_unsupported_checksum(self) -> None:
        """Unsupported ``Checksums-*`` fields are an error."""
        execute_directory = self.create_temporary_directory()
        with self.assertRaisesRegex(
            MergeUploadsError,
            "Unsupported checksum field: Checksums-Unsupported",
        ):
            self.task.merge_changes(
                execute_directory,
                [
                    deb822.Changes(
                        {
                            "Format": "1.8",
                            "Source": "hello",
                            "Version": "1.0",
                            "Checksums-Unsupported": "",
                        }
                    )
                ],
            )

    def test_merge_changes(self) -> None:
        """``merge_changes`` merges everything and symlinks referenced files."""
        download_directory = self.create_temporary_directory()

        (upload_source := download_directory / "upload-1").mkdir()
        f_dsc = upload_source / "hello_1.0.dsc"
        f_dsc.write_bytes(f_dsc.name.encode())
        f_tar = upload_source / "hello_1.0.tar.xz"
        f_tar.write_bytes(f_tar.name.encode())
        f_changes_source = upload_source / "hello_1.0_source.changes"
        ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            f_changes_source,
            [f_dsc, f_tar],
            source="hello",
            version="1.0",
        )

        (upload_all := download_directory / "upload-2").mkdir()
        f_deb = upload_all / "hello_1.0_all.deb"
        f_deb.write_bytes(f_deb.name.encode())
        f_changes_all = upload_all / "hello_1.0_all.changes"
        changes_all = ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            f_changes_all,
            [f_deb],
            source="hello",
            binaries=["hello"],
            version="1.0",
        )

        self.task._changes_paths = [f_changes_source, f_changes_all]
        execute_directory = self.create_temporary_directory()

        merged_changes_path = self.task.merge_changes(
            execute_directory,
            [read_changes(f_changes_source), read_changes(f_changes_all)],
        )

        self.assertEqual(
            read_changes(merged_changes_path),
            {
                "Architecture": "all source",
                "Binary": "hello",
                "Changes": changes_all["Changes"],
                "Checksums-Sha1": self._make_checksums_field(
                    "sha1", [f_dsc, f_tar, f_deb]
                ),
                "Checksums-Sha256": self._make_checksums_field(
                    "sha256", [f_dsc, f_tar, f_deb]
                ),
                "Date": changes_all["Date"],
                "Description": "\n hello - A Description",
                "Distribution": changes_all["Distribution"],
                "Files": self._make_files_field([f_dsc, f_tar, f_deb]),
                "Format": "1.8",
                "Maintainer": changes_all["Maintainer"],
                "Source": "hello",
                "Urgency": changes_all["Urgency"],
                "Version": "1.0",
            },
        )
        for target in (f_dsc, f_tar, f_deb):
            link = merged_changes_path.with_name(target.name)
            self.assertTrue(link.is_symlink())
            self.assertEqual(link.readlink(), target)

    def test_merge_descriptions_left_only(self) -> None:
        """Merge ``Description`` where it is only on the left side."""
        merged = deb822.Changes({"Description": "\n foo - package foo"})
        to_merge = deb822.Changes()

        self.task._merge_descriptions(merged, to_merge)

        self.assertEqual(merged["Description"], "\n foo - package foo")

    def test_merge_descriptions_right_only(self) -> None:
        """Merge ``Description`` where it is only on the right side."""
        merged = deb822.Changes()
        to_merge = deb822.Changes({"Description": "\n foo - package foo"})

        self.task._merge_descriptions(merged, to_merge)

        self.assertEqual(merged["Description"], "\n foo - package foo")

    def test_merge_descriptions_both(self) -> None:
        """Merge ``Description`` where it is on both sides."""
        merged = deb822.Changes(
            {"Description": ("\n foo - package foo\n bar - package bar")}
        )
        to_merge = deb822.Changes(
            {
                "Description": (
                    "\n"
                    " bar - package bar\n"
                    " baz - package baz\n"
                    " quux - package quux"
                )
            }
        )

        self.task._merge_descriptions(merged, to_merge)

        self.assertEqual(
            merged["Description"],
            (
                "\n"
                " foo - package foo\n"
                " bar - package bar\n"
                " baz - package baz\n"
                " quux - package quux"
            ),
        )

    def test_execute(self) -> None:
        """Test full (mocked) execution."""
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = MergeUploadsDynamicData(
            input_uploads_ids=[1, 2]
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        (upload_all := download_directory / "upload-1").mkdir()
        f_deb_all = upload_all / "meritous-data_1.5-1.1_all.deb"
        f_deb_all.write_bytes(f_deb_all.name.encode())
        f_changes_all = upload_all / "meritous_1.5-1.1_all.changes"
        changes_all = ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            f_changes_all,
            [f_deb_all],
            source="meritous",
            binaries=["meritous-data"],
            version="1.5-1.1",
        )

        (upload_amd64 := download_directory / "upload-2").mkdir()
        f_deb_lib = upload_amd64 / "libmeritous1_1.5-1.1_amd64.deb"
        f_deb_lib.write_bytes(f_deb_lib.name.encode())
        f_deb_prog = upload_amd64 / "meritous_1.5-1.1_amd64.deb"
        f_deb_prog.write_bytes(f_deb_prog.name.encode())
        f_changes_amd64 = upload_amd64 / "meritous_1.5-1.1_amd64.changes"
        ArtifactPlayground.write_deb822_file(
            deb822.Changes,
            f_changes_amd64,
            [f_deb_lib, f_deb_prog],
            source="meritous",
            binaries=["libmeritous1", "meritous"],
            version="1.5-1.1",
        )

        self.assertTrue(self.task.configure_for_execution(download_directory))
        self.assertEqual(
            self.task._changes_paths, [f_changes_all, f_changes_amd64]
        )

        execute_directory = self.create_temporary_directory()
        self.assertEqual(
            self.task.run(execute_directory), WorkRequestResults.SUCCESS
        )
        assert self.task._upload_artifact is not None
        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.SUCCESS
        )

        self.assertEqual(
            self.task._upload_artifact.data,
            DebianUpload(
                type="dpkg",
                changes_fields={
                    "Architecture": "all amd64",
                    "Binary": "libmeritous1 meritous meritous-data",
                    "Changes": changes_all["Changes"],
                    "Checksums-Sha1": self._make_checksums_field(
                        "sha1", [f_deb_all, f_deb_lib, f_deb_prog]
                    ),
                    "Checksums-Sha256": self._make_checksums_field(
                        "sha256", [f_deb_all, f_deb_lib, f_deb_prog]
                    ),
                    "Date": changes_all["Date"],
                    "Description": (
                        "\n"
                        " meritous-data - A Description\n"
                        " libmeritous1 - A Description\n"
                        " meritous - A Description"
                    ),
                    "Distribution": changes_all["Distribution"],
                    "Files": self._make_files_field(
                        [f_deb_all, f_deb_lib, f_deb_prog]
                    ),
                    "Format": "1.8",
                    "Maintainer": changes_all["Maintainer"],
                    "Source": "meritous",
                    "Urgency": changes_all["Urgency"],
                    "Version": "1.5-1.1",
                },
            ),
        )
        debusine_mock.upload_artifact.assert_called_once_with(
            Upload.create(
                changes_file=(
                    execute_directory
                    / "merged"
                    / "meritous_1.5-1.1_multi.changes"
                )
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )
        debusine_mock.relation_create.assert_has_calls(
            [
                mock.call(2, 1, "extends"),
                mock.call(2, 2, "extends"),
            ]
        )

    def test_execute_checksum_multiple_path_components(self) -> None:
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = MergeUploadsDynamicData(input_uploads_ids=[1])
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True

        (upload := download_directory / "upload-1").mkdir()
        f_changes = upload / "hello_1.0_all.changes"
        changes = deb822.Changes(
            {
                "Format": "1.8",
                "Source": "hello",
                "Version": "1.0",
                "Checksums-Sha256": "\n 0000 1 ../file",
            }
        )
        f_changes.write_text(changes.dump())

        self.assertTrue(self.task.configure_for_execution(download_directory))
        self.assertEqual(self.task._changes_paths, [f_changes])

        execute_directory = self.create_temporary_directory()
        with self.assertRaisesRegex(
            ValueError,
            r"'hello_1\.0_all\.changes' has an entry in Checksums-Sha256 with"
            r" more than one path component: '\.\./file'",
        ):
            self.task.run(execute_directory)

    def test_upload_artifacts(self) -> None:
        """upload_artifact() and relation_create() is called."""
        self.task.dynamic_data = MergeUploadsDynamicData(
            input_uploads_ids=[1, 2]
        )
        download_directory = self.create_temporary_directory()

        # Create file that will be attached when uploading the artifacts
        f_in_contents = "Format: 1.8"
        f_in1 = download_directory / "meritous_1.5-1.1_amd64.changes"
        f_in1.write_text(f_in_contents)
        f_in2 = download_directory / "meritous_1.5-1.1_all.changes"
        f_in2.write_text(f_in_contents)
        f_deb = download_directory / "meritous-data_1.5-1.1_all.deb"
        f_deb.touch()
        self.task._changes_paths = [f_in2, f_in1]
        f_merged = self.create_temporary_file()
        self.write_changes_file(f_merged, [f_deb])
        merged = read_changes(f_merged)
        merged_changes_path = (
            download_directory / "meritous_1.5-1.1_multi.changes"
        )
        with open(merged_changes_path, "w") as f:
            merged.dump(f, text_mode=True)
        self.task._upload_artifact = self.task.make_upload_artifact(
            merged_changes_path
        )

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        workspace_name = "testing"

        uploaded_artifacts = [
            create_remote_artifact(id=10, workspace=workspace_name),
        ]

        debusine_mock.upload_artifact.side_effect = uploaded_artifacts

        # self.task.workspace_name is set by the Worker
        # and is the workspace that downloads the artifact
        # containing the files needed for MergeUploads
        self.task.workspace_name = workspace_name

        # The worker set self.task.work_request_id of the task
        work_request_id = 147
        self.task.work_request_id = work_request_id

        execute_directory = self.create_temporary_directory()
        self.task.upload_artifacts(
            execute_directory, execution_result=WorkRequestResults.SUCCESS
        )

        # Assert that the artifacts were uploaded and relations created
        debusine_mock.upload_artifact.assert_called_once_with(
            self.task._upload_artifact,
            workspace=workspace_name,
            work_request=work_request_id,
        )

        # Debusine mock relation_create expected calls
        debusine_mock.relation_create.assert_has_calls(
            [
                call(uploaded_artifacts[0].id, 1, "extends"),
                call(uploaded_artifacts[0].id, 2, "extends"),
            ]
        )



# Generated by Django 5.2.7 on 2026-04-27 15:53

from django.db import migrations, models

import debusine.assets.models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0054_scoperole_observer'),
    ]

    operations = [
        migrations.RemoveConstraint(
            model_name='asset',
            name='db_asset_workspace_not_null',
        ),
        migrations.AddConstraint(
            model_name='asset',
            constraint=models.CheckConstraint(
                check=models.Q(
                    models.Q(
                        (
                            'category',
                            debusine.assets.models.AssetCategory['SIGNING_KEY'],
                        ),
                        _negated=True,
                    ),
                    ('workspace__isnull', False),
                    _connector='OR',
                ),
                name='db_asset_workspace_not_null',
            ),
        ),
        migrations.AddField(
            model_name='workerpool',
            name='apt_authentication',
            field=models.ForeignKey(
                on_delete=models.deletion.PROTECT,
                related_name='+',
                to='db.asset',
                null=True,
                blank=True,
            ),
            preserve_default=False,
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli setup command."""

import contextlib
import io
from unittest import mock

from debusine.client.cli import Cli
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.config import ConfigHandler


class CliSetupTests(BaseCliTests):
    """Tests for the setup command."""

    def assertRunsSuccessfully(self, cli: Cli) -> None:
        """Ensure cli runs successfully."""
        stderr = io.StringIO()
        with contextlib.redirect_stderr(stderr):
            cli.execute()
        self.assertEqual(stderr.getvalue(), "")

    def test_without_args(self) -> None:
        """Call setup without arguments."""
        with mock.patch("debusine.client.setup.setup_server") as setup_server:
            cli = self.create_cli(["setup"], create_config=False)
            self.assertRunsSuccessfully(cli)
        setup_server.assert_called_with(
            config_file_path=ConfigHandler.DEFAULT_CONFIG_FILE_PATH,
            server=None,
            scope=None,
        )

    def test_with_config(self) -> None:
        """Call setup with --config-file."""
        with mock.patch("debusine.client.setup.setup_server") as setup_server:
            cli = self.create_cli(
                ["--config-file=debusine.cfg", "setup"], create_config=False
            )
            self.assertRunsSuccessfully(cli)
        setup_server.assert_called_with(
            config_file_path="debusine.cfg", server=None, scope=None
        )

    def test_with_server(self) -> None:
        """Call setup with --server."""
        with mock.patch("debusine.client.setup.setup_server") as setup_server:
            cli = self.create_cli(
                ["--server=debian", "setup"], create_config=False
            )
            self.assertRunsSuccessfully(cli)
        setup_server.assert_called_with(
            config_file_path=ConfigHandler.DEFAULT_CONFIG_FILE_PATH,
            server="debian",
            scope=None,
        )

    def test_with_scope(self) -> None:
        """Call setup with --scope."""
        with mock.patch("debusine.client.setup.setup_server") as setup_server:
            cli = self.create_cli(
                ["--scope=debian", "setup"], create_config=False
            )
            self.assertRunsSuccessfully(cli)
        setup_server.assert_called_with(
            config_file_path=ConfigHandler.DEFAULT_CONFIG_FILE_PATH,
            server=None,
            scope="debian",
        )



# Generated by Django 4.2.21 on 2025-08-13 14:39

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0014_alter_workrequest_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='workspace',
            name='expiration_delay',
            field=models.DurationField(
                blank=True,
                help_text='if set, time since the last task completion time after which the workspace can be deleted',
                null=True,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the base Workflow classes."""

import abc
from typing import override
from unittest import mock

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    TaskTypes,
)
from debusine.client.models import LookupChildType
from debusine.db.models import CollectionItem, default_workspace
from debusine.db.playground import scenarios
from debusine.db.tasks import DBTask
from debusine.server.tasks.wait.models import ExternalDebsignData
from debusine.server.workflows import NoopWorkflow, Workflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import SampleWorkflow
from debusine.signing.tasks.models import DebsignData
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BaseDynamicTaskData,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
)
from debusine.test.django import TestCase, preserve_db_task_registry


class WorkflowTests(TestCase):
    """Unit tests for Workflow class."""

    scenario = scenarios.DefaultContext()

    @preserve_db_task_registry()
    def test_create(self) -> None:
        """Test instantiating a Workflow."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        wr = self.playground.create_workflow(
            task_data={},
            workspace=default_workspace(),
            task_name="exampleworkflow",
        )
        wf = wr.get_task()
        assert isinstance(wf, ExampleWorkflow)
        self.assertEqual(
            wf.data,
            BaseWorkflowData(
                task_configuration=(
                    self.scenario.default_task_configuration_collection.pk
                )
            ),
        )
        self.assertEqual(wf.work_request, wr)

    @preserve_db_task_registry()
    def test_registration(self) -> None:
        """Test class subclass registry."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        class ExampleWorkflowABC(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData],
            abc.ABC,
        ):
            """Abstract workflow subclass to use for tests."""

        class ExampleWorkflowName(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Workflow subclass with a custom name."""

            TASK_NAME = "examplename"

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        # name gets filled
        self.assertEqual(ExampleWorkflow.name, "exampleworkflow")
        self.assertEqual(ExampleWorkflowABC.name, "exampleworkflowabc")
        self.assertEqual(ExampleWorkflowName.name, "examplename")

        # Subclasses that are not ABC get registered
        self.assertIn("exampleworkflow", DBTask._sub_tasks[TaskTypes.WORKFLOW])
        self.assertNotIn(
            "exampleworkflowabc", DBTask._sub_tasks[TaskTypes.WORKFLOW]
        )
        self.assertIn("examplename", DBTask._sub_tasks[TaskTypes.WORKFLOW])

    @preserve_db_task_registry()
    def test_provides_artifact(self) -> None:
        """Test provides_artifact() update event_reactions.on_creation."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        workflow_root = self.playground.create_workflow(
            task_name="exampleworkflow"
        )
        workflow_root_task = workflow_root.get_task()
        assert isinstance(workflow_root_task, ExampleWorkflow)
        assert workflow_root.internal_collection is not None

        child = workflow_root.create_child_worker("noop")

        self.assert_work_request_event_reactions(child, on_creation=[])

        name = "hello_1:1.0-1|testing"
        data = {"architecture": "amd64"}
        category = ArtifactCategory.TEST
        artifact_filters = {"xxx": "yyy"}

        workflow_root_task.provides_artifact(
            child, category, name, data=data, artifact_filters=artifact_filters
        )

        promise = workflow_root.internal_collection.child_items.active().get(
            name=name
        )
        self.assertEqual(promise.child_type, CollectionItem.Types.BARE)
        self.assertEqual(promise.category, BareDataCategory.PROMISE)
        self.assertEqual(
            promise.data,
            {
                "promise_work_request_id": child.id,
                "promise_workflow_id": workflow_root.id,
                "promise_category": category,
                **data,
            },
        )
        on_success = [
            ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template=name,
                variables=data,
                artifact_filters={
                    **artifact_filters,
                    "category": category,
                },
            )
        ]
        self.assert_work_request_event_reactions(child, on_success=on_success)

        # provides_artifact is idempotent.
        artifact, _ = self.playground.create_artifact()
        assert workflow_root.internal_collection is not None
        workflow_root.internal_collection.manager.add_artifact(
            artifact,
            user=child.created_by,
            workflow=workflow_root,
            variables={},
            name=name,
            replace=True,
        )

        workflow_root_task.provides_artifact(
            child, category, name, data=data, artifact_filters=artifact_filters
        )

        item = workflow_root.internal_collection.child_items.active().get(
            name=name
        )
        self.assertEqual(item.child_type, CollectionItem.Types.ARTIFACT)
        self.assertEqual(item.category, ArtifactCategory.TEST)
        self.assertEqual(item.artifact, artifact)
        self.assert_work_request_event_reactions(child, on_success=on_success)

    @preserve_db_task_registry()
    def test_provides_artifact_bad_name(self) -> None:
        r"""Test requires_artifact() ValueError: name contains a slash."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        workflow_root = self.playground.create_workflow(
            task_name="exampleworkflow"
        )
        workflow_root_task = workflow_root.get_task()
        assert isinstance(workflow_root_task, Workflow)

        child = workflow_root.create_child_worker("noop")

        with self.assertRaisesRegex(
            ValueError, r'Collection item name may not contain "/"\.'
        ):
            workflow_root_task.provides_artifact(
                child, ArtifactCategory.TEST, "prefix/testing"
            )

    @preserve_db_task_registry()
    def test_provides_artifact_bad_data_key(self) -> None:
        r"""Test requires_artifact() ValueError: key starts with promise\_."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        workflow_root = self.playground.create_workflow(
            task_name="exampleworkflow"
        )
        workflow_root_task = workflow_root.get_task()
        assert isinstance(workflow_root_task, Workflow)

        child = workflow_root.create_child_worker("noop")

        key = "promise_key"
        msg = f'Field name "{key}" starting with "promise_" is not allowed.'
        with self.assertRaisesRegex(ValueError, msg):
            workflow_root_task.provides_artifact(
                child, ArtifactCategory.TEST, "testing", data={key: "value"}
            )

    @preserve_db_task_registry()
    def test_requires_artifact_lookup_single(self) -> None:
        """Test requires_artifact() call work_request.add_dependency()."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        workflow_root = self.playground.create_workflow(
            task_name="exampleworkflow"
        )
        workflow_root_task = workflow_root.get_task()
        assert isinstance(workflow_root_task, Workflow)

        # Provides the relevant artifact
        child_provides_1 = workflow_root.create_child_worker("noop")

        # Provides a non-relevant artifact (for the Lookup query of
        # the requires_artifact()
        child_provides_2 = workflow_root.create_child_worker("noop")

        child_requires = workflow_root.create_child_worker("noop")

        self.assertEqual(child_requires.dependencies.count(), 0)

        workflow_root_task.provides_artifact(
            child_provides_1, ArtifactCategory.TEST, "build-amd64"
        )
        workflow_root_task.provides_artifact(
            child_provides_2, ArtifactCategory.TEST, "build-i386"
        )
        child_provides_1.process_event_reactions("on_creation")
        child_provides_2.process_event_reactions("on_creation")

        workflow_root_task.requires_artifact(
            child_requires, "internal@collections/name:build-amd64"
        )

        self.assertEqual(child_requires.dependencies.count(), 1)
        self.assertQuerySetEqual(
            child_requires.dependencies.all(), [child_provides_1]
        )

        # Calling requires_artifact() if it was already required: is a noop
        workflow_root_task.requires_artifact(
            child_requires, "internal@collections/name:build-amd64"
        )
        self.assertEqual(child_requires.dependencies.count(), 1)

    @preserve_db_task_registry()
    def test_requires_artifact_lookup_multiple(self) -> None:
        """Test requires_artifact() call work_request.add_dependency()."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        workflow_root = self.playground.create_workflow(
            task_name="exampleworkflow"
        )
        workflow_root_task = workflow_root.get_task()
        assert isinstance(workflow_root_task, Workflow)
        assert workflow_root.internal_collection is not None

        # Providers that will match
        child_provides_1 = workflow_root.create_child_worker("noop")
        child_provides_2 = workflow_root.create_child_worker("noop")

        # Requirer
        child_requires = workflow_root.create_child_worker("noop")

        self.assertEqual(child_requires.dependencies.count(), 0)

        workflow_root_task.provides_artifact(
            child_provides_1, ArtifactCategory.TEST, "build-amd64"
        )
        workflow_root_task.provides_artifact(
            child_provides_2, ArtifactCategory.TEST, "build-i386"
        )
        child_provides_1.process_event_reactions("on_creation")
        child_provides_2.process_event_reactions("on_creation")

        # Add manually CollectionItem with a category != PROMISE for
        # unit testing coverage
        CollectionItem.objects.create_from_bare_data(
            BareDataCategory.PACKAGE_BUILD_LOG,
            parent_collection=workflow_root.internal_collection,
            name="not-relevant",
            data={},
            created_by_user=workflow_root.created_by,
            created_by_workflow=workflow_root,
        )

        workflow_root_task.requires_artifact(
            child_requires,
            LookupMultiple.model_validate(
                {
                    "collection": "internal@collections",
                    "child_type": LookupChildType.BARE,
                }
            ),
        )

        self.assertEqual(child_requires.dependencies.count(), 2)
        self.assertQuerySetEqual(
            child_requires.dependencies.all(),
            [child_provides_1, child_provides_2],
        )

    def test_lookup(self) -> None:
        """Test lookup of Workflow orchestrators."""
        self.assertEqual(
            DBTask.class_from_name(TaskTypes.WORKFLOW, "noop"), NoopWorkflow
        )
        self.assertEqual(Workflow.from_name("noop"), NoopWorkflow)

    @preserve_db_task_registry()
    def test_work_request_ensure_child(self) -> None:
        """Test work_request_ensure_child create or return work request."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        workflow_root = self.playground.create_workflow(
            task_name="exampleworkflow", task_data={}
        )

        w = workflow_root.get_task()
        assert isinstance(w, ExampleWorkflow)

        self.assertEqual(w.work_request.children.count(), 0)

        workflow_data = WorkRequestWorkflowData(
            display_name="Example step",
            step="example",
        )

        self.enterContext(
            mock.patch(
                "debusine.db.models.WorkRequest.apply_task_configuration"
            )
        )
        for method, kwargs in (
            ("worker", {"task_name": "noop", "workflow_data": workflow_data}),
            (
                "server",
                {"task_name": "servernoop", "workflow_data": workflow_data},
            ),
            (
                "internal",
                {"task_name": "workflow", "workflow_data": workflow_data},
            ),
            (
                "signing",
                {
                    "task_name": "debsign",
                    "task_data": DebsignData(unsigned=10, key="test"),
                    "workflow_data": workflow_data,
                },
            ),
            (
                "wait",
                {
                    "task_name": "externaldebsign",
                    "task_data": ExternalDebsignData(unsigned="test"),
                    "workflow_data": WorkRequestWorkflowData(
                        display_name="Example step",
                        step="example",
                        needs_input=False,
                    ),
                },
            ),
        ):
            with self.subTest(method=method, kwargs=repr(kwargs)):
                w.work_request.children.all().delete()
                factory_method = getattr(
                    w, f"work_request_ensure_child_{method}"
                )
                wr_created, created = factory_method(**kwargs)

                # One work_request got created
                self.assertEqual(w.work_request.children.count(), 1)
                self.assertTrue(created)

                # Try creating a new one (same task_name, task_data,
                # workflow_data...)
                wr_returned, created = factory_method(**kwargs)

                # No new work request created
                self.assertEqual(w.work_request.children.count(), 1)
                self.assertFalse(created)

                # Returned the same as had been created
                self.assertEqual(wr_created, wr_returned)

    @preserve_db_task_registry()
    def test_get_input_artifacts_ids(self) -> None:
        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Concrete workflow instance to use for tests."""

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        wr = self.playground.create_workflow(task_name="exampleworkflow")
        wf = wr.get_task()

        mock_artifacts = [
            InputArtifactSingle(
                artifact_id=1, lookup=1, label="source_artifact"
            ),
            InputArtifactMultiple.from_ids(
                artifact_ids=None,
                lookup=LookupMultiple(
                    ("internal@collections/name:buildlog-amd64",)
                ),
                label="package_log",
            ),
            InputArtifactMultiple.from_ids(
                artifact_ids=[5, 6],
                lookup=LookupMultiple((5, 6)),
                label="binary_artifacts",
            ),
        ]

        with mock.patch.object(
            wf, "get_input_artifacts", return_value=mock_artifacts
        ):
            self.assertEqual(wf.get_input_artifacts_ids(), [1, 5, 6])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine Web forms."""

import datetime as dt
from collections.abc import Callable
from typing import Any, TYPE_CHECKING, TypeAlias, cast, override

import yaml
from django.core.exceptions import ValidationError
from django.db.models import Q
from django.db.models.fields import BLANK_CHOICE_DASH
from django.forms import (
    BaseFormSet,
    BooleanField,
    CharField,
    CheckboxInput,
    ChoiceField,
    DateTimeField,
    Field,
    FileField,
    Form,
    IntegerField,
    ModelChoiceField,
    ModelForm,
    MultipleChoiceField,
    NumberInput,
    Textarea,
    formset_factory,
)
from django.forms.fields import InvalidJSONInput, JSONString

from debusine.artifacts.models import (
    BareDataCategory,
    CollectionCategory,
    SINGLETON_COLLECTION_CATEGORIES,
)
from debusine.db.context import context
from debusine.db.models import (
    Collection,
    Group,
    Token,
    User,
    WorkRequest,
    Workspace,
)
from debusine.db.models.collections import CollectionRelation
from debusine.db.permissioncontext import PermissionUser

if TYPE_CHECKING:
    from debusine.web.views.ui.workspaces import WorkspaceUI

    BaseFormSetBase: TypeAlias = BaseFormSet
    ModelFormBase: TypeAlias = ModelForm
    ModelChoiceFieldBase: TypeAlias = ModelChoiceField
else:
    # Django's form classes don't support generic types at run-time yet.
    from debusine.utils.typing_utils import patch_runtime_generic_support

    BaseFormSetBase = patch_runtime_generic_support(BaseFormSet)
    ModelFormBase = patch_runtime_generic_support(ModelForm)
    ModelChoiceFieldBase = patch_runtime_generic_support(ModelChoiceField)


class DictWithCallback[KT, VT](dict[KT, VT]):
    r"""
    Dictionary that when setting a value it calls a method to process it.

    Usage: d = DictWithCallback(callback, \*dict_args, \*\*dict_kwargs)

    when doing:
    d["key"] = value

    Before setting the value it calls "callback" which can change, in place,
    the value. Then sets the value to the dictionary.

    When accessing the value (d["key"]) it return the value as it was modified
    by the callback.
    """

    def __init__(
        self, callback: Callable[[VT], None], *args: Any, **kwargs: Any
    ) -> None:
        """Create the object."""
        self._callback = callback
        super().__init__(*args, **kwargs)

    @override
    def __setitem__(self, key: KT, value: VT) -> None:
        """Call self._callback(value) and set the value."""
        self._callback(value)
        super().__setitem__(key, value)


class BootstrapMixin:
    """
    Mixin that adjusts the CSS classes of form fields with Bootstrap's UI.

    This mixin is intended to be used in combination with Django's form classes.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the mixin."""
        super().__init__(*args, **kwargs)

        self._adjust_bootstrap_classes_all_fields()

        self.fields: dict[str, Field] = DictWithCallback(
            self._adjust_bootstrap_for_field, self.fields
        )

    @staticmethod
    def _adjust_bootstrap_for_field(field: Field) -> None:
        """Adjust the CSS class for a field."""
        existing_class = field.widget.attrs.get("class", "")
        bootstrap_class = None

        if field.required:
            suffix = " *"
            if not (field.label_suffix or "").endswith(suffix):
                field.label_suffix = (field.label_suffix or "") + suffix

        if isinstance(field, ChoiceField):
            bootstrap_class = "form-select"
        elif isinstance(
            field,
            (CharField, DateTimeField, DaysField, FileField, IntegerField),
        ):
            bootstrap_class = "form-control"
        elif isinstance(field, BooleanField):
            bootstrap_class = "form-check-input"

        if bootstrap_class and bootstrap_class not in existing_class.split():
            field.widget.attrs["class"] = (
                f"{existing_class} {bootstrap_class}".strip()
            )

    def _adjust_bootstrap_classes_all_fields(self) -> None:
        """Adjust the CSS classes of form fields to be Bootstrap-compatible."""
        for field in self.fields.values():
            self._adjust_bootstrap_for_field(field)


class OrderingWidget(NumberInput):
    """Local customization of the ordering widget for formsets."""

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        """Add form-control to the widget class."""
        super().__init__(attrs)
        existing_class = self.attrs.get("class", "")
        self.attrs["class"] = f"{existing_class} form-control"


class DeletionWidget(CheckboxInput):
    """Local customization of the deletion widget for formsets."""

    def __init__(self, attrs: dict[str, Any] | None = None) -> None:
        """Add form-check-input to the widget class."""
        super().__init__(attrs)
        existing_class = self.attrs.get("class", "")
        self.attrs["class"] = f"{existing_class} form-check-input"


class TokenForm(BootstrapMixin, ModelFormBase[Token]):
    """Form for creating or editing a token."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize TokenForm."""
        self.user = kwargs.pop("user")
        super().__init__(*args, **kwargs)

        if not self.instance.pk:
            # New instance (not loaded from the DB). Set defaults
            self.fields["enabled"].initial = True

    @override
    def save(self, commit: bool = True) -> Token:
        instance = super().save(commit=False)

        instance.token_type = Token.Types.USER
        instance.user = self.user

        if commit:
            instance.save()

        return instance

    class Meta:
        model = Token
        fields = ["comment", "enabled"]

        labels = {
            "comment": "Comment",
        }


class WorkspaceChoiceField(ModelChoiceFieldBase[Workspace]):
    """ChoiceField for the workspaces: set the label and order by name."""

    def __init__(
        self, user: PermissionUser | None, *args: Any, **kwargs: Any
    ) -> None:
        """Set the queryset."""
        kwargs["queryset"] = Workspace.objects.order_by("name")

        if user is None:
            # Non-authenticated users can list only public workspaces
            kwargs["queryset"] = kwargs["queryset"].filter(public=True)

        super().__init__(*args, **kwargs)

    @override
    def label_from_instance(self, obj: Workspace) -> str:
        """Return name of the workspace."""
        return obj.name


class YAMLField(CharField):
    """Edit structured data, such as a JSONField, as YAML."""

    # Note: this is mostly copied from django.forms.fields.JSONField

    default_error_messages = {
        "invalid": "Enter valid YAML data.",
    }
    widget = Textarea
    dump_kwargs: dict[str, Any] = {
        "allow_unicode": True,
        "explicit_end": False,
        "explicit_start": False,
        "sort_keys": False,
    }

    @override
    def __init__(
        self,
        encoder: Any = None,  # noqa: U100
        decoder: Any = None,  # noqa: U100
        **kwargs: Any,
    ) -> None:
        if kwargs.get("empty_value") in (dict, list):
            kwargs["empty_value"] = kwargs["empty_value"]()
        super().__init__(**kwargs)
        encoder  # fake usage for Vulture
        decoder  # fake usage for Vulture

    @override
    def to_python(self, value: Any) -> Any:
        if self.disabled:
            return value
        if value in self.empty_values:
            return self.empty_value
        elif isinstance(value, (list, dict, int, float, JSONString)):
            return value
        try:
            converted = yaml.safe_load(value)
        except yaml.YAMLError:
            raise ValidationError(
                self.error_messages["invalid"],
                code="invalid",
                params={"value": value},
            )
        if isinstance(converted, str):
            return JSONString(converted)
        else:
            return converted

    @override
    def bound_data(self, data: Any, initial: Any) -> Any:
        if self.disabled:
            return initial
        if data is None:
            return None
        try:
            return yaml.safe_load(data)
        except yaml.YAMLError:
            return InvalidJSONInput(data)

    @override
    def prepare_value(self, value: Any) -> str:
        if isinstance(value, InvalidJSONInput):
            return value
        if value in ("", None):
            value = self.empty_value
        res = yaml.safe_dump(value, **self.dump_kwargs)
        assert isinstance(res, str)
        return res

    @override
    def has_changed(self, initial: Any, data: Any) -> bool:
        if super().has_changed(initial, data):
            return True
        # For purposes of seeing whether something has changed, True isn't the
        # same as 1 and the order of keys doesn't matter.
        return yaml.safe_dump(initial, sort_keys=True) != yaml.safe_dump(
            self.to_python(data), sort_keys=True
        )


class WorkRequestUnblockForm(BootstrapMixin, Form):
    """Form for reviewing a work request awaiting manual approval."""

    # Django defaults to 10 rows, which is a bit much.  Just make it clear
    # that we accept multi-line input.
    notes = CharField(
        required=False, empty_value=None, widget=Textarea(attrs={"rows": 3})
    )


class WorkRequestConfirmForm(BootstrapMixin, Form):
    """Form for confirming a work request."""

    # Django defaults to 10 rows, which is a bit much.  Just make it clear
    # that we accept multi-line input.
    comment = CharField(
        required=False, empty_value=None, widget=Textarea(attrs={"rows": 3})
    )


class CollectionSearchForm(BootstrapMixin, Form):
    """Form for collection search fields."""

    category = ChoiceField(required=False)
    name = CharField(required=False)
    historical = BooleanField(required=False)

    def __init__(self, *args: Any, instance: Collection, **kwargs: Any) -> None:
        """Initialize category choices from the database."""
        super().__init__(*args, **kwargs)
        self.instance = instance
        choices = [("", "All")]
        choices.extend(
            (name, name)
            for name in self.instance.child_items.values_list(
                "category", flat=True
            )
            .distinct()
            .order_by("category")
        )
        cast(ChoiceField, self.fields["category"]).choices = choices


class WorkflowFilterForm(Form):
    """Form for filtering workflows."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize form."""
        super().__init__(*args, **kwargs)

        statuses: list[tuple[str, str | list[tuple[str, str]]]] = []
        for status_value, status_label in WorkRequest.Statuses.choices:
            if status_value == WorkRequest.Statuses.RUNNING:
                running_options: list[tuple[str, str]] = []
                for (
                    runtime_value,
                    runtime_label,
                ) in WorkRequest.RuntimeStatuses.choices:
                    running_options.append(
                        (
                            f"running__{runtime_value}",
                            f"Running (children: {runtime_label})",
                        )
                    )

                # Add "Any" for any Running status.
                running_options.append(
                    ("running__any", "Running (children: Any)")
                )
                statuses.append(("Running", running_options))
            else:
                statuses.append((status_value, status_label))

        self.fields["statuses"] = MultipleChoiceField(
            choices=statuses,
            required=False,
        )

        runtime_statuses_choices = [
            (status.value, status.label)
            for status in WorkRequest.RuntimeStatuses
        ]

        self.fields["runtime_statuses"] = MultipleChoiceField(
            choices=runtime_statuses_choices,
            required=False,
        )

        self.fields["results"] = MultipleChoiceField(
            choices=sorted(
                choice for choice in WorkRequest.Results.choices if choice[0]
            ),
            required=False,
        )

        self.fields["with_failed_work_requests"] = BooleanField(
            required=False,
        )


class DaysField(IntegerField):
    """Edit a timedelta as a number of days."""

    @override
    def prepare_value(self, value: Any) -> Any:
        """Convert a timedelta to days."""
        if isinstance(value, dt.timedelta):
            return super().prepare_value(value.days)
        return super().prepare_value(value)

    @override
    def to_python(self, value: Any) -> Any:
        """Convert days to timedelta."""
        res = super().to_python(value)
        if res is not None:
            return dt.timedelta(days=res)
        return res


class WorkspaceForm(BootstrapMixin, ModelFormBase[Workspace]):
    """Form for configuring a workspace."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Set help texts for the fields."""
        super().__init__(*args, **kwargs)

        self.fields["expiration_delay"].help_text = (
            "If unset, workspace is permanent. Otherwise, the workspace gets "
            "removed after the given period of inactivity (expressed as a "
            "number of days after the completion of the last work request)"
        )
        self.fields["default_expiration_delay"].help_text = (
            "Number of days that new artifacts and work requests are kept "
            "in the workspace before being expired (0 means no expiration)"
        )

    class Meta:
        model = Workspace
        fields = ["public", "expiration_delay", "default_expiration_delay"]
        field_classes = {
            "expiration_delay": DaysField,
            "default_expiration_delay": DaysField,
        }


class WorkspaceInheritanceForm(BootstrapMixin, Form):
    """Form for one item in the inheritance chain of a workspace."""

    parent = ChoiceField()

    def __init__(
        self, workspace_ui: "WorkspaceUI", *args: Any, **kwargs: Any
    ) -> None:
        """Set help texts for the fields."""
        super().__init__(*args, **kwargs)
        self.workspace_ui = workspace_ui
        cast(ChoiceField, self.fields["parent"]).choices = [(None, "None")] + [
            (ws.pk, str(ws)) for ws in self.workspace_ui.candidate_parents
        ]


class WorkspaceInheritanceFormSet(BaseFormSetBase[WorkspaceInheritanceForm]):
    """FormSet to edit an inheritance chain."""

    ordering_widget = OrderingWidget
    deletion_widget = DeletionWidget

    def __init__(
        self, *args: Any, workspace_ui: "WorkspaceUI", **kwargs: Any
    ) -> None:
        """Store a WorkspaceUI to use to fill in form choices."""
        super().__init__(*args, **kwargs)
        self.workspace_ui = workspace_ui

    @override
    def get_form_kwargs(self, index: int | None) -> dict[str, Any]:
        """Pass the WorkspaceUI to use to fill in form choices."""
        kwargs = super().get_form_kwargs(index)
        kwargs["workspace_ui"] = self.workspace_ui
        return kwargs

    @classmethod
    def formset_factory(
        cls, workspace_ui: "WorkspaceUI"
    ) -> type[BaseFormSetBase[WorkspaceInheritanceForm]]:
        """Create a Form class for a WorkspaceUI."""

        class Partial(WorkspaceInheritanceFormSet):
            def __init__(
                self,
                *args: Any,
                **kwargs: Any,
            ) -> None:
                """Store a WorkspaceUI to use to fill in form choices."""
                super().__init__(*args, workspace_ui=workspace_ui, **kwargs)

        return formset_factory(
            WorkspaceInheritanceForm,
            formset=Partial,
            can_order=True,
            can_delete=True,
            extra=1,
        )


class GroupAddUserForm(BootstrapMixin, Form):
    """Form for adding a user to a group."""

    username = CharField(required=True)

    def __init__(self, group: Group, *args: Any, **kwargs: Any) -> None:
        """Set the group the user would be added to."""
        self.group = group
        super().__init__(*args, **kwargs)

    @override
    def clean(self) -> dict[str, Any]:
        """
        Validate username.

        The user needs to exist and not be already a member of the group.
        """
        cleaned_data = super().clean()
        assert cleaned_data is not None

        if "username" not in cleaned_data:
            return cleaned_data

        try:
            user = User.objects.get(username=cleaned_data["username"])
        except User.DoesNotExist:
            self.add_error(
                "username", f"User {cleaned_data['username']} does not exist"
            )
            return cleaned_data

        if self.group.users.filter(pk=user.pk).exists():
            self.add_error(
                "username", f"User is already a member of {self.group}"
            )

        return cleaned_data


class TaskConfigurationInspectorForm(BootstrapMixin, Form):
    """Form for looking up task configurations."""

    task = ChoiceField()
    subject = ChoiceField(required=False)
    context = ChoiceField(required=False)
    tags_provided = CharField(
        required=False,
        help_text="space-separated list of task tags a task provides",
    )
    tags_required = CharField(
        required=False,
        help_text="space-separated list of worker tags a task requires",
    )

    def __init__(
        self, *args: Any, collection: Collection, **kwargs: Any
    ) -> None:
        """Initialize choices from the database."""
        super().__init__(*args, **kwargs)
        children = collection.child_items.filter(
            Q(data__template=None) | ~Q(data__has_key="template"),
            category=BareDataCategory.TASK_CONFIGURATION,
        )

        task_choices = []
        for ttype, tname in (
            children.order_by("data__task_type", "data__task_name")
            .values_list("data__task_type", "data__task_name")
            .distinct()
        ):
            task_choices.append((f"{ttype}:{tname}", f"{ttype}:{tname}"))
        cast(ChoiceField, self.fields["task"]).choices = task_choices

        subject_choices = [("", "(unspecified)")] + [
            (subject, subject)
            for subject in (
                children.order_by("data__subject")
                .values_list("data__subject", flat=True)
                .distinct()
            )
            if subject is not None
        ]
        cast(ChoiceField, self.fields["subject"]).choices = subject_choices

        context_choices = [("", "(unspecified)")] + [
            (child_context, child_context)
            for child_context in (
                children.order_by("data__context")
                .values_list("data__context", flat=True)
                .distinct()
            )
            if child_context is not None
        ]
        cast(ChoiceField, self.fields["context"]).choices = context_choices


class CollectionFormBase(BootstrapMixin, ModelFormBase[Collection]):
    """Common functions for collection forms."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize object."""
        super().__init__(*args, **kwargs)
        # Add measurement unit to interval fields
        for field_name in (
            "full_history_retention_period",
            "metadata_only_retention_period",
        ):
            if field := self.fields.get(field_name):
                field.help_text = field.help_text + " (in days)"

    def check_name(
        self,
        workspace: Workspace,
        name: str,
        category: CollectionCategory,
        instance: Collection | None = None,
    ) -> None:
        """Check that the name field matches collection constraints."""
        conflicts = Collection.objects.filter(
            name=name, workspace=workspace, category=category
        )
        if instance is not None:
            conflicts = conflicts.exclude(pk=instance.pk)
        if conflicts.exists():
            self.add_error(
                "name",
                f"A collection called {name}@{category}"
                f" already exists in workspace {workspace}",
            )

        if category in SINGLETON_COLLECTION_CATEGORIES and name != "_":
            self.add_error(
                "name", 'A singleton collection needs to be named "_"'
            )
        if category not in SINGLETON_COLLECTION_CATEGORIES and name.startswith(
            "_"
        ):
            self.add_error(
                "name",
                'The name of a non-singleton collection cannot start with "_"',
            )


class CollectionUpdateForm(CollectionFormBase):
    """Form for configuring a collection."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize object."""
        super().__init__(*args, **kwargs)
        # Avoid conflicting with Form.data
        self.fields["data"] = YAMLField(required=False, empty_value=dict)

    @override
    def clean(self) -> dict[str, Any]:
        """Enforce collection name constraints."""
        cleaned_data = super().clean()
        assert cleaned_data is not None
        self.check_name(
            self.instance.workspace,
            cleaned_data["name"],
            CollectionCategory(self.instance.category),
            instance=self.instance,
        )
        return cleaned_data

    class Meta:
        model = Collection
        fields = [
            "name",
            "full_history_retention_period",
            "metadata_only_retention_period",
            "data",
        ]
        field_classes = {
            "full_history_retention_period": DaysField,
            "metadata_only_retention_period": DaysField,
        }


class CollectionCreateForm(CollectionFormBase):
    """Form for creating a collection."""

    category = ChoiceField()

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize object."""
        super().__init__(*args, **kwargs)
        # Avoid conflicting with Form.data
        self.fields["data"] = YAMLField(required=False, empty_value=dict)

        assert isinstance(self.fields["category"], ChoiceField)
        self.fields["category"].choices = [
            (c, c)
            for c in sorted(CollectionCategory)
            if c != CollectionCategory.WORKFLOW_INTERNAL
        ]

    @override
    def clean(self) -> dict[str, Any]:
        """Enforce collection name constraints."""
        cleaned_data = super().clean()
        assert cleaned_data is not None
        if "name" in cleaned_data and "category" in cleaned_data:
            self.check_name(
                context.require_workspace(),
                str(cleaned_data["name"]),
                CollectionCategory(cleaned_data["category"]),
            )
        return cleaned_data

    @override
    def save(self, commit: bool = True) -> Collection:
        instance = super().save(commit=False)
        instance.workspace = context.require_workspace()
        instance.retains_artifacts = Collection.RetainsArtifacts.ALWAYS
        if commit:
            instance.save()
        return instance

    class Meta:
        model = Collection
        fields = [
            "name",
            "category",
            "full_history_retention_period",
            "metadata_only_retention_period",
            "data",
        ]
        field_classes = {
            "full_history_retention_period": DaysField,
            "metadata_only_retention_period": DaysField,
        }


class CollectionRelationCreateForm(
    BootstrapMixin, ModelFormBase[CollectionRelation]
):
    """Form for creating a collection relation."""

    source_collection: Collection

    def __init__(
        self,
        *args: Any,
        source_collection: Collection,
        allowed_types: set[CollectionRelation.Types],
        **kwargs: Any,
    ) -> None:
        """Initialize form."""
        super().__init__(*args, **kwargs)

        self.source_collection = source_collection

        # Relation targets sorted by name
        target_field = self.fields["target"]
        assert hasattr(target_field, "queryset")
        target_field.queryset = target_field.queryset.order_by("name")

        # Relation types: only the possible ones are displayed (and sorted)
        type_field = self.fields["type"]
        assert hasattr(type_field, "choices")

        self.fields[
            "target"
        ].help_text = "Only targets allowed for the selected type are shown"

        # Remove the type_fields that are not allowed and also the blank choice
        type_field.choices = sorted(
            [
                (value, label)
                for value, label in type_field.choices
                if (value, label) not in BLANK_CHOICE_DASH
                and value in allowed_types
            ]
        )

        type_field.help_text = (
            "Some types may be hidden if they have no available targets"
        )

    class Meta:
        model = CollectionRelation
        fields = ["type", "target"]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Move task_configuration_id to configured_task_data."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps


def move_to_configured_task_data(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Move task_configuration_id to configured_task_data."""
    WorkRequest = apps.get_model("db", "WorkRequest")

    for wr in WorkRequest.objects.filter(
        dynamic_task_data__has_key="task_configuration_id"
    ):
        value = wr.dynamic_task_data.pop("task_configuration_id")
        if value is not None:
            configured_task_data = wr.configured_task_data or {}
            if configured_task_data.get("task_configuration") is None:
                configured_task_data["task_configuration"] = value
                wr.configured_task_data = configured_task_data
        wr.save()


def move_to_dynamic_task_data(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Move task_configuration_id back to dynamic_task_data."""
    WorkRequest = apps.get_model("db", "WorkRequest")

    for wr in WorkRequest.objects.filter(
        configured_task_data__has_key="task_configuration"
    ):
        value = wr.configured_task_data.pop("task_configuration")
        if value is not None:
            dynamic_task_data = wr.dynamic_task_data or {}
            dynamic_task_data["task_configuration_id"] = value
            wr.dynamic_task_data = dynamic_task_data
        wr.save()


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0019_workflowtemplate_db_workflowtemplate_update_suites'),
    ]

    operations = [
        migrations.RunPython(
            move_to_configured_task_data, move_to_dynamic_task_data
        ),
    ]



# Generated by Django 4.2.20 on 2025-04-29 10:39

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0001_squashed_0_11_0'),
    ]
    replaces = [
        ('db', '0143_workrequest_aborted_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='aborted_by',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='aborted_work_requests',
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
View extension to model sidebar elements.

Sidebar elements are informational elements or actions that can be used to
populate the UI sidebar.
"""

import datetime as dt
from dataclasses import dataclass
from functools import cached_property
from typing import override

from django.db.models import Model
from django.template.context import BaseContext
from django.utils.formats import date_format
from django.utils.html import format_html
from django.utils.safestring import SafeString
from django.utils.timesince import timesince

from debusine.db.models import Artifact, User, WorkRequest, Worker, Workspace
from debusine.web.icons import Icons
from debusine.web.views.base import Widget


@dataclass(kw_only=True, frozen=True)
class SidebarItem(Widget):
    """Renderable element for the right sidebar."""

    #: icon to use
    icon: str = "app"
    #: label describing what the field is about
    label: str
    #: user-presented value for the field
    value: str | None = None
    #: detailed information about the value
    detail: str | None = None

    @cached_property
    def content(self) -> str:
        """Select the main content to display."""
        if self.value:
            return self.value
        if self.detail:
            return self.detail
        return "-"

    @cached_property
    def main_tooltip(self) -> str:
        """Select what to use as tooltip for the whole entry."""
        if self.detail and self.value and self.detail != self.value:
            return self.detail
        return self.label

    @cached_property
    def icon_tooltip(self) -> str | None:
        """Select if a tooltip is shown on the icon, and what it is."""
        if self.main_tooltip == self.label:
            return None
        return self.label

    @override
    def render(self, context: BaseContext) -> str:
        raise NotImplementedError(
            f"{self.__class__.__name__}.render not implemented"
        )


@dataclass(kw_only=True, frozen=True)
class SidebarElement(SidebarItem):
    """Informative element for the right sidebar."""

    @override
    def render(self, context: BaseContext) -> str:  # noqa: ARG002, U100
        if self.icon_tooltip:
            icon_title = format_html(' title="{}"', self.icon_tooltip)
        else:
            icon_title = SafeString()
        return format_html(
            '<span class="list-group-item" title="{main_tooltip}">'
            '<span class="bi bi-{icon}"{icon_title}></span> '
            '{content}</span>',
            main_tooltip=self.main_tooltip,
            icon_title=icon_title,
            icon=self.icon,
            content=self.content,
        )


@dataclass(kw_only=True, frozen=True)
class SidebarAction(SidebarItem):
    """Action element for the right sidebar."""

    # Action URL
    url: str

    @override
    def render(self, context: BaseContext) -> str:  # noqa: ARG002, U100
        if self.icon_tooltip:
            icon_title = format_html(' title="{}"', self.icon_tooltip)
        else:
            icon_title = SafeString()
        return format_html(
            '<a class="list-group-item list-group-item-action'
            ' d-flex justify-content-between"'
            ' title="{main_tooltip}" href="{url}">'
            '<span><span class="bi bi-{icon}"{icon_title}></span> '
            '{content}</span>'
            '<span>…</span>'
            '</a>',
            main_tooltip=self.main_tooltip,
            icon_title=icon_title,
            url=self.url,
            icon=self.icon,
            content=self.content,
        )


def create_workspace(workspace: Workspace) -> SidebarItem:
    """Create workspace information."""
    return SidebarAction(
        url=workspace.get_absolute_url(),
        icon="person-workspace",
        label="Workspace",
        value=workspace.name,
    )


def create_user(
    user: User | None, *, context: Model | None = None
) -> SidebarItem:
    """Create user information."""
    match context:
        case WorkRequest():
            title = "User who created the work request"
        case Artifact():
            title = "User who created the artifact"
        case _:
            title = "User"

    return SidebarElement(
        icon=Icons.USER,
        label=title,
        value=str(user) if user else None,
    )


def create_created_at(at: dt.datetime) -> SidebarItem:
    """Create created_at information."""
    return SidebarElement(
        icon=Icons.CREATED_AT,
        label="Created",
        value=timesince(at) + " ago",
        detail=date_format(at, "DATETIME_FORMAT"),
    )


def create_expire_at(at: dt.datetime | None) -> SidebarItem:
    """Create expire_at information."""
    kwargs = {
        "icon": Icons.EXPIRE_AT,
        "label": "Expiration date",
    }
    if at:
        return SidebarElement(
            detail=date_format(at, "DATETIME_FORMAT"),
            value=timesince(at, reversed=True),
            **kwargs,
        )
    else:
        return SidebarElement(value="Never", **kwargs)


def create_worker(worker: Worker | None) -> SidebarItem:
    """Create a link to a worker."""
    kwargs = {
        "icon": Icons.WORKER,
        "label": "Worker",
    }
    if worker is None:
        return SidebarElement(**kwargs)
    return SidebarAction(
        value=worker.name, url=worker.get_absolute_url(), **kwargs
    )



# Generated by Django 4.2.28 on 2026-05-24 07:44

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0058_non_numeric_usernames'),
    ]

    operations = [
        migrations.AddField(
            model_name='workflowtemplate',
            name='extra_groups',
            field=models.ManyToManyField(
                help_text='Extra groups granted when running a workflow',
                related_name='granted_by_workflow_template',
                to='db.group',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Infrastructure to create test scenarios in the database."""

import base64
import contextlib
import datetime as dt
import tempfile
import textwrap
import time as tm
import traceback
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path
from secrets import token_hex
from typing import Any, Literal, TYPE_CHECKING, assert_never, overload
from unittest import mock

from django.conf import settings
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.db.models import F, Model
from django.test import override_settings
from django.utils import timezone

from debusine.artifacts import LocalArtifact
from debusine.artifacts.local_artifact import BinaryPackage, SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    ArtifactData,
    BareDataCategory,
    BaseArtifactDataModel,
    CollectionCategory,
    DebianBinaryPackage,
    DebianPackageBuildLog,
    DebianSourcePackage,
    DebusineSigningInput,
    DebusineSigningOutput,
    SigningResult,
    TaskTypes,
)
from debusine.artifacts.playground import ArtifactPlayground
from debusine.assets import (
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
    AssetCategory,
    BaseAssetDataModel,
    BasicAPTAuthenticationData,
    BasicAPTCredentials,
    CloudProviderAccountData,
    CloudProvidersType,
    DummyProviderAccountData,
    HetznerProviderAccountConfiguration,
    HetznerProviderAccountCredentials,
    HetznerProviderAccountData,
    KeyPurpose,
    SigningKeyData,
)
from debusine.client.models import RuntimeParameter, RuntimeParameters
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Asset,
    AssetUsage,
    Collection,
    CollectionItem,
    DEFAULT_FILE_STORE_NAME,
    File,
    FileInArtifact,
    FileStore,
    FileUpload,
    Group,
    Scope,
    Token,
    User,
    WellKnownWorkflowTemplate,
    WorkRequest,
    Worker,
    WorkerPool,
    WorkflowTemplate,
    Workspace,
)
from debusine.db.models.permissions import (
    get_resource_scope,
    get_resource_workspace,
    resolve_role,
)
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.db.tests.utils import _calculate_hash_from_data
from debusine.server.collections.lookup import lookup_single
from debusine.server.file_backend.interface import FileBackendInterface
from debusine.server.worker_pools import (
    AWSEC2WorkerPoolSpecification,
    DummyWorkerPoolSpecification,
    HetznerCloudWorkerPoolSpecification,
    WorkerPoolLimits,
    WorkerPoolSpecifications,
)
from debusine.server.workflows.base import (
    Workflow,
    WorkflowRunError,
    orchestrate_workflow,
)
from debusine.server.workflows.models import (
    BaseWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BackendType,
    BaseTaskData,
    SbuildBuildComponent,
    SbuildData,
    SbuildInput,
    WorkerType,
)

if TYPE_CHECKING:
    from debusine.db.playground.scenarios import Scenario


class DEFAULT_WORKSPACE:
    """
    Sentinel object for methods that may have NULL workspaces.

    Replace with PEP-661 sentinels with python 3.15.
    """


@dataclass(kw_only=True)
class UploadArtifacts:
    """The return value for Playground.create_upload_artifacts()."""

    upload: Artifact
    source: Artifact | None
    binaries: list[Artifact] | None


@dataclass(kw_only=True)
class SourceUploadArtifacts(UploadArtifacts):
    """UploadArtifacts for a source-only upload."""

    source: Artifact
    binaries: None


@dataclass(kw_only=True)
class BinaryUploadArtifacts(UploadArtifacts):
    """UploadArtifacts for a binary-only upload."""

    source: None
    binaries: list[Artifact]


@dataclass(kw_only=True)
class MixedUploadArtifacts(UploadArtifacts):
    """UploadArtifacts for a mixed upload."""

    source: Artifact
    binaries: list[Artifact]


class Playground(ArtifactPlayground):
    """Generate data scenarios for the database."""

    def __init__(
        self,
        default_file_store_name: str = DEFAULT_FILE_STORE_NAME,
        default_workspace_name: str = settings.DEBUSINE_DEFAULT_WORKSPACE,
        default_username: str = "playground",
        default_user_password: str | None = None,
        default_user_email: str = "playground@example.org",
        default_user_first_name: str = "Play",
        default_user_last_name: str = "Ground",
    ) -> None:
        """Set default values."""
        super().__init__()
        self.default_file_store_name = default_file_store_name
        self.default_workspace_name = default_workspace_name
        self.default_username = default_username
        self.default_user_password = default_user_password
        self.default_user_email = default_user_email
        self.default_user_first_name = default_user_first_name
        self.default_user_last_name = default_user_last_name
        # Registry of scenarios created by this playground.
        #
        # This can be used when instantiating multiple scenarios, allowing one
        # to access previously built ones
        self.scenarios: dict[str, Scenario] = {}
        self._seq_create_pc_with_extra_groups = 0

    @context.disable_permission_checks()
    def get_default_user(self) -> User:
        """Return the default user for playground methods."""
        try:
            return User.objects.get(username=self.default_username)
        except User.DoesNotExist:
            user = User.objects.create_user(
                username=self.default_username,
                email=self.default_user_email,
                first_name=self.default_user_first_name,
                last_name=self.default_user_last_name,
            )
            if self.default_user_password is not None:
                user.set_password(self.default_user_password)
                user.save()
        return user

    @context.disable_permission_checks()
    def get_default_file_store(self) -> FileStore:
        """Get the default file store used."""
        try:
            return FileStore.objects.get(name=self.default_file_store_name)
        except FileStore.DoesNotExist:
            return FileStore.objects.create(
                name=self.default_file_store_name,
                backend=FileStore.BackendChoices.LOCAL,
                configuration={},
            )

    def get_default_scope(self) -> Scope:
        """Get the default workspace used by create methods."""
        # TODO: when tests pass with scoped workspaces, switch to a differently
        # named default scope for tests, to catch code possibly depending on
        # the fallback scope introduced by the migration
        # return self.get_or_create_scope(name="tests")
        return self.get_or_create_scope(name=settings.DEBUSINE_DEFAULT_SCOPE)

    @context.disable_permission_checks()
    def get_default_workspace(self) -> Workspace:
        """Get the default workspace used by create methods."""
        return self.create_workspace(public=True)

    @context.disable_permission_checks()
    def add_user_permission(
        self, user: User, model: type[Model], codename: str
    ) -> None:
        """Add a permission to a user."""
        user.user_permissions.add(
            Permission.objects.get(
                codename=codename,
                content_type=ContentType.objects.get_for_model(model),
            )
        )

    @context.disable_permission_checks()
    def get_or_create_scope(self, name: str, **kwargs: Any) -> Scope:
        """Create a scope, or return an existing scope."""
        kwargs.setdefault("label", name.capitalize())
        scope, created = Scope.objects.get_or_create(name=name, defaults=kwargs)
        if created:
            scope.file_stores.add(self.get_default_file_store())
        return scope

    @context.disable_permission_checks()
    def create_user(
        self,
        username: str,
    ) -> User:
        """Create a user."""
        email = f"{username}@example.org"
        user = User.objects.create_user(
            username=username,
            email=email,
        )
        return user

    @context.disable_permission_checks()
    def create_group(
        self,
        name: str,
        users: list[User] | None = None,
        scope: Scope | None = None,
        workspace: Workspace | None = None,
        ephemeral: bool = False,
    ) -> Group:
        """Create a group, setting membership to users."""
        if scope is None:
            scope = self.get_default_scope()
        if users is None:
            users = []
        group, _ = Group.objects.get_or_create(
            scope=scope, workspace=workspace, name=name, ephemeral=ephemeral
        )
        group.users.set(users)
        return group

    @context.disable_permission_checks()
    def create_group_role(
        self,
        resource: Model,
        role: str,
        users: list[User] | None = None,
        name: str | None = None,
    ) -> Group:
        """Create a group with the given role on a resource."""
        if name is None:
            if isinstance(resource, Workspace):
                name = role
            else:
                name = f"{str(resource).replace('/', '-')}-{role}"
        if users is None:
            users = []
        role = resolve_role(resource, role)
        group = self.create_group(
            name,
            users,
            workspace=get_resource_workspace(resource),
            scope=get_resource_scope(resource),
        )
        roles_model = getattr(resource.__class__, "objects").get_roles_model()
        roles_model.objects.filter(group=group).delete()
        group.assign_role(resource, role)
        return group

    @contextlib.contextmanager
    def assign_role(
        self, resource: Model, user: PermissionUser, *roles: str
    ) -> Generator[None]:
        """Temporarily assign one or more roles to the user on the resource."""
        groups: list[Group] = []

        if roles:
            assert user.is_authenticated
            for role in roles:
                name = f"test_assign_role_{role}_{tm.monotonic_ns()}"
                group = self.create_group_role(
                    resource, role, [user], name=name
                )
                groups.append(group)

        try:
            yield
        finally:
            for group in groups:
                group.delete()

    @context.disable_permission_checks()
    def add_user(
        self, group: Group, user: User, role: Group.Roles = Group.Roles.MEMBER
    ) -> None:
        """Add a user to a group."""
        group.add_user(user, role)

    @context.disable_permission_checks()
    @transaction.atomic
    def create_worker_token(
        self,
        *,
        worker: Worker | None = None,
        enabled: bool = True,
        set_worker: bool = True,
        **kwargs: Any,
    ) -> Token:
        """
        Return a worker token.

        Optionally create the worker if none is supplied.
        """
        token = Token.objects.create(token_type=Token.Types.WORKER, **kwargs)

        if set_worker:
            if worker is None:
                worker = self.create_worker(token=token)
            worker.token = token
            worker.save()
            token.refresh_from_db()
            assert token._worker_id == worker.id

        if enabled:
            token.enable()
        else:
            token.disable()
        return token

    @context.disable_permission_checks()
    @transaction.atomic
    def create_worker_activation_token(
        self,
        *,
        worker: Worker | None = None,
        enabled: bool = True,
        set_worker: bool = True,
        **kwargs: Any,
    ) -> Token:
        """
        Return a worker activation token.

        Optionally create the worker if none is supplied.
        """
        token = Token.objects.create(
            token_type=Token.Types.WORKER_ACTIVATION, **kwargs
        )

        if set_worker:
            if worker is None:
                worker = self.create_worker(activation_token=token)
            worker.activation_token = token
            worker.save()
            token.refresh_from_db()
            assert token._activating_worker_id == worker.id

        if enabled:
            token.enable()
        else:
            token.disable()
        return token

    @context.disable_permission_checks()
    @transaction.atomic
    def create_work_request_token(
        self,
        work_request: WorkRequest | None = None,
        worker: Worker | None = None,
    ) -> Token:
        """
        Return a worker's limited work request token.

        Optionally create the work request if none is supplied.
        Optionally create the worker if none is supplied.
        """
        if work_request is None:
            if worker is None:
                worker = self.create_worker()
            work_request = self.create_worker_task(
                mark_running=True, worker=worker
            )

        token = Token.objects.create_work_request_token(work_request)
        # Refresh token, since it may have been updated by triggers
        token.refresh_from_db()
        return token

    @context.disable_permission_checks()
    def create_user_token(
        self, user: User | None = None, enabled: bool = True, **kwargs: Any
    ) -> Token:
        """
        Return a user token.

        :param user: user to associate, defaulting to get_default_user()
        """
        if user is None:
            user = self.get_default_user()
        token = Token.objects.create(
            token_type=Token.Types.USER, user=user, **kwargs
        )
        if enabled:
            token.enable()
        else:
            token.disable()
        return token

    @context.disable_permission_checks()
    def create_external_user_token(
        self, *, username: str = "test", enabled: bool = True, **kwargs: Any
    ) -> Token:
        """Return an external user token."""
        token = Token.objects.create_external_user_token(
            username=username, **kwargs
        )
        if enabled:
            token.enable()
        else:
            token.disable()
        return token

    @context.disable_permission_checks()
    def create_workspace(
        self, *, scope: Scope | None = None, **kwargs: Any
    ) -> Workspace:
        """Create a Workspace and return it."""
        if scope is None:
            scope = self.get_default_scope()
        workspace_name = kwargs.pop("name", self.default_workspace_name)
        workspace, _ = Workspace.objects.select_related("scope").get_or_create(
            scope=scope, name=workspace_name, defaults=kwargs
        )

        return workspace

    @context.disable_permission_checks()
    def create_parent_workspace(
        self, *, child: Workspace | None = None, **kwargs: Any
    ) -> Workspace:
        """
        Create a new workspace as parent of an existing workspace.

        If the child workspace is not specified, then add it to the
        inheritance chain of the default workspace.
        """
        if child is None:
            child = self.get_default_workspace()
        assert "name" in kwargs

        parent = self.create_workspace(**kwargs)
        child.inherits.add(
            parent, through_defaults={"order": child.inherits.count() + 1}
        )

        return parent

    @context.disable_permission_checks()
    def create_collection(
        self,
        name: str,
        category: CollectionCategory,
        *,
        workspace: Workspace | None = None,
        data: dict[str, Any] | None = None,
    ) -> Collection:
        """Create a collection."""
        return Collection.objects.create(
            name=name,
            category=category,
            workspace=workspace or self.get_default_workspace(),
            data=data or {},
        )

    @context.disable_permission_checks()
    def create_singleton_collection(
        self,
        category: CollectionCategory,
        *,
        workspace: Workspace | None = None,
        data: dict[str, Any] | None = None,
    ) -> Collection:
        """Create a singleton collection."""
        ret, _ = Collection.objects.get_or_create_singleton(
            category=category,
            workspace=workspace or self.get_default_workspace(),
            data=data or {},
        )
        return ret

    @context.disable_permission_checks()
    def create_workflow_template(
        self,
        name: str,
        task_name: str,
        *,
        workspace: Workspace | None = None,
        static_parameters: dict[str, Any] | None = None,
        runtime_parameters: RuntimeParameters = RuntimeParameter.ANY,
        **kwargs: Any,
    ) -> WorkflowTemplate:
        """Create a workflow template."""
        return WorkflowTemplate.objects.create(
            name=name,
            workspace=workspace or self.get_default_workspace(),
            task_name=task_name,
            static_parameters=static_parameters or {},
            runtime_parameters=runtime_parameters,
            **kwargs,
        )

    @context.disable_permission_checks()
    def create_workflow(
        self,
        task_name: str | WorkflowTemplate = "noop",
        task_data: BaseWorkflowData | dict[str, Any] | None = None,
        *,
        parent: WorkRequest | None = None,
        created_by: User | None = None,
        workspace: Workspace | None = None,
        validate: bool = True,
        mark_pending: bool = True,
        mark_running: bool = False,
        mark_aborted: bool = False,
        result: WorkRequest.Results | None = None,
        internal_collection_artifacts: (
            list[tuple[Artifact, dict[str, Any]]] | None
        ) = None,
    ) -> WorkRequest:
        """
        Create a workflow.

        :param mark_running: if True call mark_running() method
        :param mark_aborted: if True call mark_aborted() method
        :param result: if not None call mark_completed(result)
        :param internal_collection_artifacts: list of tuples with the artifacts
           and arguments for the add_artifact() method. Adds the artifacts
           with the same created_by user as the workflow
        """
        if created_by is None:
            created_by = self.get_default_user()

        if isinstance(task_name, WorkflowTemplate):
            template = task_name
            task_name = template.task_name
            if workspace is None:
                workspace = template.workspace
        else:
            workflow_template_name = getattr(
                WellKnownWorkflowTemplate,
                task_name.upper(),
                f"{task_name}-template",
            )
            templates = WorkflowTemplate.objects.filter(
                name=workflow_template_name
            )
            if workspace is not None:
                templates = templates.filter(workspace=workspace)
            try:
                template = templates.get()
                assert template.task_name == task_name
                workspace = template.workspace
            except WorkflowTemplate.DoesNotExist:
                template = None
                if workspace is None:
                    workspace = self.get_default_workspace()

        if isinstance(task_data, BaseWorkflowData):
            task_data = task_data.model_dump(mode="json", exclude_unset=True)

        with contextlib.ExitStack() as stack:
            stack.enter_context(context.local())
            context.reset()
            context.set_permission_context_scoped(
                workspace.scope, PermissionContext.for_ui_user(created_by)
            )
            workspace.set_current()
            stack.enter_context(context.disable_permission_checks())
            if not validate:
                stack.enter_context(
                    mock.patch(
                        "debusine.db.models.WorkRequest._validate_workflow"
                    )
                )
                stack.enter_context(
                    mock.patch("debusine.db.models.WorkRequest.full_clean")
                )
                stack.enter_context(
                    mock.patch(
                        "debusine.db.models.WorkRequest"
                        ".apply_task_configuration"
                    )
                )
            work_request = WorkRequest.objects.create_workflow(
                template=template,
                workspace=workspace,
                task_name=task_name,
                data=task_data or {},
                parent=parent,
            )
            if internal_collection_artifacts is not None:
                assert work_request.internal_collection is not None

                for artifact, add_args in internal_collection_artifacts:
                    work_request.internal_collection.manager.add_artifact(
                        artifact, user=created_by, **add_args
                    )

            self.advance_work_request(
                work_request,
                mark_pending=mark_pending,
                mark_running=mark_running,
                mark_aborted=mark_aborted,
                result=result,
            )
            return work_request

    def _create_worker_for_work_request(
        self, work_request: WorkRequest
    ) -> Worker:
        """Create a new worker that can run the given work request."""
        match work_request.task_type:
            case TaskTypes.WORKER:
                return self.create_worker()
            case TaskTypes.SERVER:
                return self.create_worker(worker_type=WorkerType.CELERY)
            case TaskTypes.SIGNING:
                return self.create_worker(worker_type=WorkerType.SIGNING)
            case _:
                raise NotImplementedError(
                    "Cannot create a worker that can run"
                    f" a {work_request.task_type} work request"
                )

    @contextlib.contextmanager
    def wrap_task_status_changes(
        self, use_data: bool = True
    ) -> Generator[None]:
        """
        Wrap code that changes task status, applying mocking if needed.

        :param use_data: if True, change task status normally. If False, mock
          functions that would use lookups in task data so that they succeed
          eve if task_data contains invalid lookups
        """
        with contextlib.ExitStack() as stack:
            if not use_data:

                def mock_apply_task_configuration(self_: WorkRequest) -> None:
                    self_.configured_task_data = self_.task_data

                def mock_compute_dynamic_data(self_: WorkRequest) -> None:
                    pass

                stack.enter_context(
                    mock.patch(
                        "debusine.db.models.WorkRequest"
                        ".apply_task_configuration",
                        side_effect=mock_apply_task_configuration,
                        autospec=True,
                    )
                )
                stack.enter_context(
                    mock.patch(
                        "debusine.db.models.WorkRequest.compute_dynamic_data",
                        side_effect=mock_compute_dynamic_data,
                        autospec=True,
                    )
                )
            yield

    # Complexity is too high, but postponing refactoring this for now, as it
    # can have a fallout in the entire test suite
    def advance_work_request(  # noqa: C901
        self,
        work_request: WorkRequest,
        *,
        mark_pending: bool = False,
        mark_running: bool = False,
        mark_aborted: bool = False,
        result: WorkRequest.Results | None = None,
        completed_at: dt.datetime | None = None,
        worker: Worker | None = None,
        assign_new_worker: bool = False,
        use_data: bool = True,
    ) -> None:
        """
        Postprocess a newly created WorkRequest, simulating its lifetime.

        :param mark_pending: if True call mark_pending() method
        :param mark_running: if True call mark_running() method (it implies
          mark_pending and assign_new_worker)
        :param mark_aborted: if True call mark_aborted() method
        :param result: if not None call mark_completed(result)
        :param completed_at: if not None, time of completion
        :param assign_new_worker: if True assign worker to the work request (it
          implies mark_pending)
        :param use_data: by default, go through task configuration and dynamic
          data. If set to False, mock them so the task data doesn't need to be
          consistent with the database
        """
        assert not (worker and assign_new_worker)
        if worker is None and assign_new_worker:
            worker = self._create_worker_for_work_request(work_request)
        del assign_new_worker

        if result is not None:
            mark_running = True
        if mark_running:
            mark_pending = True
            if not worker and (
                work_request.task_type
                in (TaskTypes.WORKER, TaskTypes.SERVER, TaskTypes.SIGNING)
            ):
                worker = self._create_worker_for_work_request(work_request)
        if worker:
            mark_pending = True

        with self.wrap_task_status_changes(use_data=use_data):
            if (
                mark_pending
                and work_request.status == WorkRequest.Statuses.BLOCKED
            ):
                work_request.mark_pending()

            if worker and work_request.worker is None:
                work_request.assign_worker(worker)

            if (
                mark_running
                and work_request.status == WorkRequest.Statuses.PENDING
            ):
                work_request.mark_running()

            if mark_aborted:
                work_request.mark_aborted()

            if result is not None:
                work_request.mark_completed(result)

            if completed_at is not None:
                work_request.completed_at = completed_at
                work_request.save()

    def orchestrate_workflow(
        self, work_request: WorkRequest, use_data: bool = True
    ) -> bool:
        """
        Orchestrate a workflow.

        Errors in work requests are printed instead of logged.

        :param use_data: if False, mock things so that invalid lookups in task
          data are tolerated.
        """

        def _log_workflow_run_error(exc: WorkflowRunError) -> None:
            """Print WorkflowRunError exceptions instead of logging them."""
            traceback.print_exception(exc)

        with (
            self.wrap_task_status_changes(use_data=use_data),
            mock.patch(
                "debusine.server.workflows.base._log_workflow_run_error",
                side_effect=_log_workflow_run_error,
            ),
        ):
            return orchestrate_workflow(work_request)

    def orchestrate_child(
        self, parent: Workflow[Any, Any], child: WorkRequest
    ) -> None:
        """
        Call ``parent.orchestrate_child(child)``.

        Errors in work requests are printed instead of logged.
        """

        def _log_workflow_run_error(
            exc: WorkflowRunError,
        ) -> None:  # pragma: no cover
            """Print WorkflowRunError exceptions instead of logging them."""
            traceback.print_exception(exc)

        with (
            mock.patch(
                "debusine.server.workflows.base._log_workflow_run_error",
                side_effect=_log_workflow_run_error,
            ),
        ):
            parent.orchestrate_child(child)

    def schedule_for_worker(self, worker: Worker) -> WorkRequest | None:
        """
        Run scheduling for a worker.

        Exceptions raise by the WorkRequest during assignment are re-raised
        instead of logged: do not use this method when testing for handling of
        expected failures.

        :returns: the work request assigned to the worker (or None)
        """
        # Prevent import loop
        from debusine.server.scheduler import schedule_for_worker

        def _log_work_request_assign_error(
            _: WorkRequest, exc: Exception
        ) -> None:
            """Re-raise exceptions instead of logging them."""
            raise exc

        with mock.patch(
            "debusine.server.scheduler._log_work_request_assign_error",
            side_effect=_log_work_request_assign_error,
        ):
            return schedule_for_worker(worker, set())

    @context.disable_permission_checks()
    def create_worker_task(
        self,
        mark_pending: bool = False,
        mark_running: bool = False,
        mark_aborted: bool = False,
        result: WorkRequest.Results | None = None,
        assign_new_worker: bool = False,
        expired: bool = False,
        workspace: Workspace | None = None,
        task_name: str = "noop",
        task_data: BaseTaskData | dict[str, Any] | None = None,
        assign_contributor_role: bool = False,
        completed_at: dt.datetime | None = None,
        validate: bool = True,
        worker: Worker | None = None,
        use_data: bool = True,
        **kwargs: Any,
    ) -> WorkRequest:
        """
        Return a new instance of WorkRequest, for a worker task.

        :param mark_running: if True call mark_running() method
        :param mark_aborted: if True call mark_aborted() method
        :param assign_new_worker: if True assign worker to the work request
        :param result: if not None call mark_completed(result)
        :param expired: True to create an expired work request
          (created_at: 1 year ago, expiration_delay: 1 day)
        :param assign_contributor_role: assign the ``CONTRIBUTOR`` role on
          the workspace to the creator of the work request
        :param completed_at: if present, set completed_at after creation
        :param use_data: by default, go through task configuration and dynamic
          data. If set to False, mock them so the task data so that invalid
          lookups are tolerated
        :param kwargs: use them when creating the WorkRequest model
        """
        assert "task_type" not in kwargs

        # FileStore and Workspace are usually created by the migrations
        # When using a TransactionTestCase with async methods
        # the tests don't have access to the created FileStore / Workspace
        # yet (need verification, but they seem to be in a non-committed
        # transaction and the test code is in a different thread because
        # of the implementation of database_sync_to_async
        if workspace is None:
            workspace = self.create_workspace(public=True)

        defaults: dict[str, Any] = {
            "task_name": task_name,
            "workspace": workspace,
            "task_data": task_data,
        }
        if "created_by" not in kwargs:
            defaults["created_by"] = self.get_default_user()
        if expired:
            defaults.update(
                created_at=timezone.now() - dt.timedelta(days=365),
                expiration_delay=dt.timedelta(days=1),
            )

        defaults.update(kwargs)

        created_at = defaults.pop("created_at", None)

        with contextlib.ExitStack() as stack:
            if not validate:
                stack.enter_context(
                    mock.patch("debusine.db.models.WorkRequest.full_clean")
                )
            work_request = WorkRequest.objects.create_worker(**defaults)

        # created_at is set to now automatically by create via auto_now_add, so
        # if a specific value was requested we need to set it after creation
        if created_at is not None:
            work_request.created_at = created_at
            work_request.save()

        if assign_contributor_role:
            self.create_group_role(
                work_request.workspace,
                Workspace.Roles.CONTRIBUTOR,
                users=[work_request.created_by],
            )

        self.advance_work_request(
            work_request,
            mark_pending=mark_pending,
            mark_running=mark_running,
            mark_aborted=mark_aborted,
            result=result,
            completed_at=completed_at,
            worker=worker,
            assign_new_worker=assign_new_worker,
            use_data=use_data,
        )

        return work_request

    @context.disable_permission_checks()
    def create_server_task(
        self,
        task_name: str = "servernoop",
        task_data: BaseTaskData | dict[str, Any] | None = None,
        mark_pending: bool = False,
        mark_running: bool = False,
        mark_aborted: bool = False,
        worker: Worker | None = None,
        result: WorkRequest.Results | None = None,
    ) -> WorkRequest:
        """Return a new instance of WorkRequest, for a server task."""
        work_request = WorkRequest.objects.create_server(
            task_name=task_name,
            task_data=task_data,
            created_by=self.get_default_user(),
            workspace=self.create_workspace(public=True),
        )
        self.advance_work_request(
            work_request,
            mark_pending=mark_pending,
            mark_running=mark_running,
            mark_aborted=mark_aborted,
            result=result,
            worker=worker,
        )
        return work_request

    @context.disable_permission_checks()
    def create_internal_task(
        self,
        task_name: str = "noop",
        mark_pending: bool = False,
        mark_running: bool = False,
        mark_aborted: bool = False,
        result: WorkRequest.Results | None = None,
        validate: bool = True,
    ) -> WorkRequest:
        """Return a new instance of WorkRequest, for an internal task."""
        with contextlib.ExitStack() as stack:
            if not validate:
                stack.enter_context(
                    mock.patch("debusine.db.models.WorkRequest.full_clean")
                )
            work_request = WorkRequest.objects.create_internal(
                task_name=task_name,
                workspace=self.create_workspace(public=True),
                created_by=self.get_default_user(),
            )
            self.advance_work_request(
                work_request,
                mark_pending=mark_pending,
                mark_running=mark_running,
                mark_aborted=mark_aborted,
                result=result,
            )
            return work_request

    @context.disable_permission_checks()
    def create_signing_task(
        self,
        mark_pending: bool = False,
        mark_running: bool = False,
        mark_aborted: bool = False,
        result: WorkRequest.Results | None = None,
        workspace: Workspace | None = None,
        task_name: str = "noop",
        task_data: BaseTaskData | dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> WorkRequest:
        """
        Return a new instance of WorkRequest, for a signing task.

        :param mark_running: if True call mark_running() method
        :param kwargs: use them when creating the WorkRequest model
        """
        assert "task_type" not in kwargs

        # FileStore and Workspace are usually created by the migrations
        # When using a TransactionTestCase with async methods
        # the tests don't have access to the created FileStore / Workspace
        # yet (need verification, but they seem to be in a non-committed
        # transaction and the test code is in a different thread because
        # of the implementation of database_sync_to_async
        if workspace is None:
            workspace = self.create_workspace(public=True)

        defaults: dict[str, Any] = {
            "task_name": task_name,
            "workspace": workspace,
        }
        if "created_by" not in kwargs:
            defaults["created_by"] = self.get_default_user()
        if isinstance(task_data, BaseTaskData):
            defaults["task_data"] = task_data.model_dump(
                mode="json", exclude_unset=True
            )
        else:
            defaults["task_data"] = task_data or {}

        defaults.update(kwargs)

        work_request = WorkRequest.objects.create_signing(**defaults)

        self.advance_work_request(
            work_request,
            mark_pending=mark_pending,
            mark_running=mark_running,
            mark_aborted=mark_aborted,
            result=result,
        )

        return work_request

    @context.disable_permission_checks()
    def create_wait_task(
        self,
        *,
        task_name: str = "noop",
        task_data: BaseTaskData | dict[str, Any] | None = None,
        mark_pending: bool = False,
        mark_running: bool = False,
        mark_aborted: bool = False,
        result: WorkRequest.Results | None = None,
        **kwargs: Any,
    ) -> WorkRequest:
        """
        Return a new instance of WorkRequest, for a wait task.

        :param mark_running: if True call mark_running() method
        :param result: if not None call mark_completed(result)
        :param kwargs: use them when creating the WorkRequest model
        """
        assert "task_type" not in kwargs

        defaults: dict[str, Any] = {
            "task_name": task_name,
            "workspace": self.create_workspace(public=True),
            "task_data": task_data,
        }
        if "created_by" not in kwargs:
            defaults["created_by"] = self.get_default_user()
        defaults.update(kwargs)

        work_request = WorkRequest.objects.create_wait(**defaults)

        self.advance_work_request(
            work_request,
            mark_pending=mark_pending,
            mark_running=mark_running,
            mark_aborted=mark_aborted,
            result=result,
        )

        return work_request

    @context.disable_permission_checks()
    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @transaction.atomic
    def create_worker(
        self,
        worker_type: WorkerType = WorkerType.EXTERNAL,
        extra_dynamic_metadata: dict[str, Any] | None = None,
        worker_pool: WorkerPool | None = None,
        fqdn: str = "computer.lan",
        token: Token | None = None,
        activation_token: Token | None = None,
    ) -> Worker:
        """
        Return a new Worker.

        :param concurrency: if set, set the worker concurrency
        """
        if worker_pool is not None:
            Worker.objects.create_pool_members(worker_pool=worker_pool, count=1)
            worker = Worker.objects.filter(
                worker_type=worker_type,
                worker_pool=worker_pool,
            ).latest("registered_at", "id")
        elif activation_token is not None:
            assert token is None
            worker = Worker.objects.create_with_fqdn(
                fqdn, worker_type=worker_type, activation_token=activation_token
            )
            # Refresh token objects, since they may have been updated by
            # triggers.
            assert worker.activation_token is not None
            worker.activation_token.refresh_from_db()
            activation_token.refresh_from_db()
        else:
            worker = Worker.objects.create_with_fqdn(
                fqdn,
                worker_type=worker_type,
                token=token or self.create_worker_token(set_worker=False),
            )
            # Refresh token objects, since they may have been updated by
            # triggers.
            assert worker.token is not None
            worker.token.refresh_from_db()
            if token is not None:
                token.refresh_from_db()

        dynamic_metadata = {
            "system:cpu_cores": 4,
            "system:worker_type": worker_type,
            "sbuild:version": 1,
        }

        if extra_dynamic_metadata:
            dynamic_metadata.update(extra_dynamic_metadata)

        worker.set_dynamic_metadata(dynamic_metadata)

        if worker_pool:
            worker.instance_created_at = timezone.now()
            worker.save()

        return worker

    @context.disable_permission_checks()
    def create_worker_pool(
        self,
        name: str = "test",
        enabled: bool = True,
        architectures: list[str] | None = None,
        tags: list[str] | None = None,
        provider_account: Asset | None = None,
        specifications: WorkerPoolSpecifications | None = None,
        instance_wide: bool = True,
        ephemeral: bool = False,
        limits: WorkerPoolLimits | None = None,
        apt_authentication: Asset | None = None,
    ) -> WorkerPool:
        """Return a new WorkerPool."""
        if architectures is None:
            architectures = ["amd64", "i386"]
        if tags is None:
            tags = []
        if provider_account is None:
            provider_account = self.create_cloud_provider_account_asset()
        if specifications is None:
            provider_type = provider_account.data["provider_type"]
            match provider_type:
                case CloudProvidersType.DUMMY:
                    specifications = DummyWorkerPoolSpecification()
                case CloudProvidersType.AWS:
                    specifications = AWSEC2WorkerPoolSpecification(
                        launch_templates=[]
                    )
                case CloudProvidersType.HETZNER:
                    specifications = HetznerCloudWorkerPoolSpecification(
                        server_type="cx22",
                        image_name="debian-13",
                    )
                case _:
                    raise NotImplementedError(f"No support for {provider_type}")
        if limits is None:
            limits = WorkerPoolLimits()
        pool = WorkerPool.objects.create(
            name=name,
            provider_account=provider_account,
            enabled=enabled,
            architectures=architectures,
            tags=tags,
            specifications=specifications.model_dump(),
            instance_wide=instance_wide,
            ephemeral=ephemeral,
            limits=limits.model_dump(),
            registered_at=timezone.now(),
            apt_authentication=apt_authentication,
        )
        return pool

    @context.disable_permission_checks()
    def create_file(self, contents: bytes = b"test") -> File:
        """
        Create a File model and return the saved fileobj.

        :param contents: used to compute hash digest and size
        """
        hashed = _calculate_hash_from_data(contents)
        file, _ = File.objects.get_or_create(
            hash_digest=hashed, size=len(contents)
        )
        return file

    @context.disable_permission_checks()
    def create_file_in_backend(
        self,
        backend: FileBackendInterface[Any] | None = None,
        contents: bytes = b"test",
    ) -> File:
        """
        Create a temporary file and adds it in the backend.

        :param backend: file backend to add the file in
        :param contents: contents of the file
        """
        if backend is None:
            backend = self.get_default_file_store().get_backend_object()
        with tempfile.NamedTemporaryFile("w+b") as fd:
            fd.write(contents)
            fd.flush()

            return backend.add_file(Path(fd.name))

    @context.disable_permission_checks()
    def create_bare_data_item(
        self,
        parent_collection: Collection,
        name: str,
        *,
        category: BareDataCategory = BareDataCategory.TEST,
        data: BaseArtifactDataModel | dict[str, Any] | None = None,
        created_by_user: User | None = None,
        created_by_workflow: WorkRequest | None = None,
    ) -> CollectionItem:
        """Create a collection item holding bare data."""
        return CollectionItem.objects.create_from_bare_data(
            parent_collection=parent_collection,
            name=name,
            category=category,
            data=data or {},
            created_by_user=created_by_user or self.get_default_user(),
            created_by_workflow=created_by_workflow,
        )

    @context.disable_permission_checks()
    def create_artifact(
        self,
        paths: list[str] | dict[str, bytes] | None = None,
        files_size: int = 100,
        *,
        category: ArtifactCategory = ArtifactCategory.TEST,
        workspace: Workspace | None = None,
        data: ArtifactData | dict[str, Any] | None = None,
        expiration_delay: int | None = None,
        work_request: WorkRequest | None = None,
        created_by: User | None = None,
        create_files: bool = False,
        skip_add_files_in_store: bool = False,
    ) -> tuple[Artifact, dict[str, bytes]]:
        """
        Create an artifact and return tuple with the artifact and files data.

        :param paths: list of paths to create (will contain random data)
        :param files_size: size of the test data
        :param category: this artifact's category (see
          :ref:`artifact-reference`)
        :param data: key-value data for this artifact (see
          :ref:`artifact-reference`)
        :param expiration_delay: set expiration_delay field (in days)
        :param work_request: work request that created this artifact
        :param created_by: set Artifact.created_by to it
        :param create_files: create a file and add it into the LocalFileBackend
        :param skip_add_files_in_store: do not add the files in the store
          (only create the File object in the database)

        This method return a tuple:
        - artifact: Artifact
        - files_contents: Dict[str, bytes] (paths and test data)
        """
        # Import here to avoid a circular loop
        from debusine.test.test_utils import data_generator

        if skip_add_files_in_store and not create_files:
            raise ValueError(
                "skip_add_files_in_store must be False if create_files is False"
            )

        if workspace is None:
            workspace = (
                work_request.workspace
                if work_request is not None
                else self.get_default_workspace()
            )

        match data:
            case ArtifactData():
                raw_data = data.model_dump(mode="json")
            case dict():
                raw_data = data
            case None:
                raw_data = {}
            case _ as unreachable:
                assert_never(unreachable)

        artifact = Artifact.objects.create(
            category=category,
            workspace=workspace,
            data=raw_data,
            expiration_delay=(
                dt.timedelta(expiration_delay)
                if expiration_delay is not None
                else None
            ),
            created_by_work_request=work_request,
            created_by=created_by,
        )

        data_gen = data_generator(files_size)

        files_contents = {}
        if isinstance(paths, dict):
            files_contents.update(paths)
        elif paths is None:
            pass
        else:
            for path in paths:
                files_contents[path] = next(data_gen)

        if create_files:
            file_backend = workspace.scope.file_stores.order_by(
                F("filestoreinscope__upload_priority").desc(nulls_last=True)
            )[0].get_backend_object()

            for path, contents in files_contents.items():
                if skip_add_files_in_store:
                    fileobj = self.create_file(contents)
                else:
                    fileobj = self.create_file_in_backend(
                        file_backend, contents
                    )

                FileInArtifact.objects.create(
                    artifact=artifact,
                    path=path,
                    file=fileobj,
                    complete=not skip_add_files_in_store,
                )

        return artifact, files_contents

    @context.disable_permission_checks()
    def create_artifact_from_local(
        self,
        local_artifact: LocalArtifact[Any],
        *,
        workspace: Workspace | None = None,
        expiration_delay: int | None = None,
        work_request: WorkRequest | None = None,
        created_by: User | None = None,
        create_files: bool = False,
        skip_add_files_in_store: bool = False,
    ) -> Artifact:
        """Create an artifact in the database from a `LocalArtifact`."""
        artifact, _ = self.create_artifact(
            paths={
                name: path.read_bytes()
                for name, path in local_artifact.files.items()
            },
            category=local_artifact.category,
            workspace=workspace or self.get_default_workspace(),
            data=local_artifact.data,
            expiration_delay=expiration_delay,
            work_request=work_request,
            created_by=created_by or self.get_default_user(),
            create_files=create_files,
            skip_add_files_in_store=skip_add_files_in_store,
        )
        return artifact

    @context.disable_permission_checks()
    def create_file_upload(
        self, *, workspace: Workspace | None = None
    ) -> FileUpload:
        """
        Create a new FileUpload object.

        Create the workspace, artifact, file and file_in_artifact associated
        to the file_upload object.
        """
        artifact, _ = self.create_artifact(
            paths=["README"],
            workspace=workspace,
            create_files=True,
            skip_add_files_in_store=True,
        )

        file_in_artifact = artifact.fileinartifact_set.first()
        assert file_in_artifact is not None

        return FileUpload.objects.create(
            file_in_artifact=file_in_artifact,
            path=f"temp_file_{tm.monotonic_ns()}",
        )

    @context.disable_permission_checks()
    def create_artifact_relation(
        self,
        artifact: Artifact,
        target: Artifact,
        relation_type: ArtifactRelation.Relations = (
            ArtifactRelation.Relations.RELATES_TO
        ),
    ) -> ArtifactRelation:
        """Create an ArtifactRelation."""
        return ArtifactRelation.objects.create(
            artifact=artifact, target=target, type=relation_type
        )

    @context.disable_permission_checks()
    def create_source_artifact(
        self,
        *,
        name: str = "hello",
        binaries: list[str] | None = None,
        version: str = "1.0-1",
        architectures: set[str] | None = None,
        workspace: Workspace | None = None,
        work_request: WorkRequest | None = None,
        created_by: User | None = None,
        create_files: bool = False,
    ) -> Artifact:
        """Create an artifact for a source package."""
        with tempfile.TemporaryDirectory() as tempdir:
            workdir = Path(tempdir)
            source_package = self.create_source_package(
                workdir,
                name=name,
                binaries=binaries,
                version=version,
                architectures=architectures,
            )
            return self.create_artifact_from_local(
                source_package,
                workspace=workspace,
                work_request=work_request,
                created_by=created_by,
                create_files=create_files,
            )

    def create_minimal_source_package_artifact(
        self,
        name: str = "hello",
        version: str = "1.0-1",
        *,
        paths: list[str] | dict[str, bytes] | None = None,
        create_files: bool = False,
        dsc_fields: dict[str, Any] | None = None,
    ) -> Artifact:
        """Create a minimal `debian:source-package` artifact."""
        dsc_fields = {"Package": name, "Version": version, **(dsc_fields or {})}
        artifact, _ = self.create_artifact(
            category=ArtifactCategory.SOURCE_PACKAGE,
            data=DebianSourcePackage(
                name=name, version=version, type="dpkg", dsc_fields=dsc_fields
            ),
            paths=paths,
            create_files=create_files,
        )
        return artifact

    def create_minimal_binary_packages_artifact(
        self,
        srcpkg_name: str = "hello",
        srcpkg_version: str = "1.0-1",
        version: str = "1.0-1",
        architecture: str = "all",
        *,
        paths: list[str] | dict[str, bytes] | None = None,
        create_files: bool = False,
    ) -> Artifact:
        """Create a minimal `debian:binary-packages` artifact."""
        artifact, _ = self.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGES,
            data={
                "srcpkg_name": srcpkg_name,
                "srcpkg_version": srcpkg_version,
                "version": version,
                "architecture": architecture,
                "packages": [],
            },
            paths=paths,
            create_files=create_files,
        )
        return artifact

    def create_minimal_binary_package_artifact(
        self,
        srcpkg_name: str = "hello",
        srcpkg_version: str = "1.0-1",
        package: str = "hello",
        version: str = "1.0-1",
        architecture: str = "all",
        *,
        paths: list[str] | dict[str, bytes] | None = None,
        workspace: Workspace | None = None,
        create_files: bool = False,
        deb_fields: dict[str, Any] | None = None,
    ) -> Artifact:
        """Create a minimal `debian:binary-package` artifact."""
        deb_fields = {
            "Package": package,
            "Version": version,
            "Architecture": architecture,
            **(deb_fields or {}),
        }
        artifact, _ = self.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data=DebianBinaryPackage(
                srcpkg_name=srcpkg_name,
                srcpkg_version=srcpkg_version,
                deb_fields=deb_fields,
                deb_control_files=[],
            ),
            paths=paths,
            workspace=workspace,
            create_files=create_files,
        )
        return artifact

    @overload
    def create_upload_artifacts(
        self,
        *,
        source: Literal[True] = True,
        src_name: str = ...,
        version: str = ...,
        src_architectures: set[str] | None = ...,
        binary: Literal[False],
        binaries: list[tuple[str, str]] | None = ...,
        section: str = ...,
        priority: str = ...,
        workspace: Workspace | None = ...,
        work_request: WorkRequest | None = None,
        created_by: User | None = ...,
        create_files: bool = ...,
    ) -> SourceUploadArtifacts: ...

    @overload
    def create_upload_artifacts(
        self,
        *,
        source: Literal[False],
        src_name: str = ...,
        version: str = ...,
        src_architectures: set[str] | None = ...,
        binary: Literal[True] = True,
        binaries: list[tuple[str, str]] | None = ...,
        section: str = ...,
        priority: str = ...,
        workspace: Workspace | None = ...,
        work_request: WorkRequest | None = None,
        created_by: User | None = ...,
        create_files: bool = ...,
    ) -> BinaryUploadArtifacts: ...

    @overload
    def create_upload_artifacts(
        self,
        *,
        source: Literal[True] = True,
        src_name: str = ...,
        version: str = ...,
        src_architectures: set[str] | None = ...,
        binary: Literal[True] = True,
        binaries: list[tuple[str, str]] | None = ...,
        section: str = ...,
        priority: str = ...,
        workspace: Workspace | None = ...,
        work_request: WorkRequest | None = None,
        created_by: User | None = ...,
        create_files: bool = ...,
    ) -> MixedUploadArtifacts: ...

    @context.disable_permission_checks()
    def create_upload_artifacts(
        self,
        *,
        source: bool = True,
        src_name: str = "hello",
        version: str = "1.0-1",
        src_architectures: set[str] | None = None,
        binary: bool = True,
        binaries: list[tuple[str, str]] | None = None,
        section: str = "unknown",
        priority: str = "optional",
        workspace: Workspace | None = None,
        work_request: WorkRequest | None = None,
        created_by: User | None = None,
        create_files: bool = False,
    ) -> UploadArtifacts:
        """Create a set of artifacts for a package upload."""
        if binaries is None:
            binaries = [(src_name, "all")]
        local_artifacts: dict[str, list[LocalArtifact[Any]]] = {
            "source": [],
            "binaries": [],
        }
        artifacts: dict[str, list[Artifact]] = {}
        with tempfile.TemporaryDirectory() as tempdir:
            workdir = Path(tempdir)
            upload = self.create_upload(
                workdir,
                source=source,
                src_name=src_name,
                version=version,
                src_architectures=src_architectures,
                binary=binary,
                binaries=binaries,
                section=section,
                priority=priority,
            )
            local_artifacts["upload"] = [upload]
            if source:
                source_package = SourcePackage.create(
                    name=src_name,
                    version=version,
                    files=[
                        file
                        for file in upload.files.values()
                        if file.name.endswith((".dsc", ".tar.gz", ".tar.xz"))
                    ],
                )
                local_artifacts["source"] = [source_package]
            if binary:
                binary_packages: list[LocalArtifact[Any]] = [
                    BinaryPackage.create(file=file)
                    for file in upload.files.values()
                    if file.suffix == ".deb"
                ]
                local_artifacts["binaries"] = binary_packages

            for key, values in local_artifacts.items():
                artifacts[key] = [
                    self.create_artifact_from_local(
                        local_artifact,
                        workspace=workspace,
                        work_request=work_request,
                        created_by=created_by,
                        create_files=create_files,
                    )
                    for local_artifact in values
                ]

        if source:
            self.create_artifact_relation(
                artifact=artifacts["upload"][0],
                target=artifacts["source"][0],
                relation_type=ArtifactRelation.Relations.EXTENDS,
            )
        for binary_artifact in artifacts["binaries"]:
            self.create_artifact_relation(
                artifact=artifacts["upload"][0],
                target=binary_artifact,
                relation_type=ArtifactRelation.Relations.EXTENDS,
            )

        return UploadArtifacts(
            upload=artifacts["upload"][0],
            source=artifacts["source"][0] if source else None,
            binaries=artifacts["binaries"] or None,
        )

    @context.disable_permission_checks()
    def create_build_log_artifact(
        self,
        *,
        source: str = "hello",
        version: str = "1.0-1",
        build_arch: str = "amd64",
        workspace: Workspace | None = None,
        work_request: WorkRequest | None = None,
        created_by: User | None = None,
        contents: bytes | None = None,
        skip_add_files_in_store: bool = False,
    ) -> Artifact:
        """Create an artifact for a build log."""
        filename = f"{source}_{version}_{build_arch}.buildlog"
        if contents is None:
            contents = "\n".join(
                f"Line {lineno} of {filename}" for lineno in range(1, 11)
            ).encode()
        if created_by is None:
            if work_request is not None:
                created_by = work_request.created_by
            else:
                created_by = self.get_default_user()
        data = DebianPackageBuildLog(
            source=source,
            version=version,
            architecture=build_arch,
            filename=filename,
        )
        artifact, _ = self.create_artifact(
            paths={filename: contents},
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            workspace=workspace or self.get_default_workspace(),
            data=data.model_dump(),
            work_request=work_request,
            created_by=created_by,
            create_files=True,
            skip_add_files_in_store=skip_add_files_in_store,
        )
        return artifact

    def create_signing_input_artifact(
        self,
        binary_package_name: str | None = "hello",
        trusted_certs: list[str] | None = None,
        workspace: Workspace | None = None,
        work_request: WorkRequest | None = None,
    ) -> Artifact:
        """Create a minimal `debusine:signing-input` artifact."""
        artifact, _ = self.create_artifact(
            category=ArtifactCategory.SIGNING_INPUT,
            data=DebusineSigningInput(
                trusted_certs=trusted_certs,
                binary_package_name=binary_package_name,
            ),
            workspace=workspace,
            work_request=work_request,
        )
        return artifact

    def create_signing_output_artifact(
        self,
        purpose: KeyPurpose,
        fingerprint: str,
        results: list[SigningResult] | None = None,
        binary_package_name: str | None = None,
        workspace: Workspace | None = None,
        work_request: WorkRequest | None = None,
    ) -> Artifact:
        """Create a minimal `debusine:signing-output` artifact."""
        artifact, _ = self.create_artifact(
            category=ArtifactCategory.SIGNING_OUTPUT,
            data=DebusineSigningOutput(
                purpose=purpose,
                fingerprint=fingerprint,
                results=results or [],
                binary_package_name=binary_package_name,
            ),
            workspace=workspace,
            work_request=work_request,
        )
        return artifact

    def create_repository_index(
        self,
        path: str,
        contents: bytes | None = None,
        workspace: Workspace | None = None,
        work_request: WorkRequest | None = None,
        skip_add_files_in_store: bool = False,
    ) -> Artifact:
        """Create a minimal ``debian:repository-index`` artifact."""
        artifact, _ = self.create_artifact(
            paths=[path] if contents is None else {path: contents},
            category=ArtifactCategory.REPOSITORY_INDEX,
            workspace=workspace,
            work_request=work_request,
            create_files=True,
            skip_add_files_in_store=skip_add_files_in_store,
        )
        return artifact

    @context.disable_permission_checks()
    def create_asset(
        self,
        category: AssetCategory,
        data: BaseAssetDataModel | dict[str, Any],
        workspace: Workspace | None = None,
        created_by: User | None = None,
        created_by_work_request: WorkRequest | None = None,
    ) -> Asset:
        """Create an Asset."""
        if created_by is None:
            created_by = self.get_default_user()
        return Asset.objects.create(
            category=category,
            workspace=workspace,
            data=(
                data.model_dump()
                if isinstance(data, BaseAssetDataModel)
                else data
            ),
            created_by=created_by,
            created_by_work_request=created_by_work_request,
        )

    def create_signing_key_asset(
        self,
        purpose: KeyPurpose = KeyPurpose.OPENPGP_UPLOAD,
        fingerprint: str | None = None,
        public_key: str | None = None,
        description: str | None = "Test Key",
        workspace: Workspace | None = None,
        created_by: User | None = None,
        created_by_work_request: WorkRequest | None = None,
    ) -> Asset:
        """Create a debusine:signing-key asset."""
        if fingerprint is None:
            fingerprint = token_hex()
        if public_key is None:
            public_key = textwrap.dedent(
                f"""\
            -----BEGIN PGP PUBLIC KEY BLOCK-----

            Comment: This isn't a valid key.

            {token_hex()}
            {token_hex()}
            {token_hex()}

            -----END PGP PUBLIC KEY BLOCK-----
            """
            )
        if workspace is None:
            workspace = self.get_default_workspace()
        return self.create_asset(
            category=AssetCategory.SIGNING_KEY,
            data=SigningKeyData(
                purpose=purpose,
                fingerprint=fingerprint,
                public_key=base64.b64encode(public_key.encode()).decode(),
                description=description,
            ),
            workspace=workspace,
            created_by=created_by,
            created_by_work_request=created_by_work_request,
        )

    @context.disable_permission_checks()
    def create_cloud_provider_account_asset(
        self,
        data: CloudProviderAccountData | None = None,
        cloud_provider: CloudProvidersType | None = None,
        name: str | None = None,
        workspace: Workspace | None = None,
        created_by: User | None = None,
        created_by_work_request: WorkRequest | None = None,
    ) -> Asset:
        """Create a debusine:cloud-provider-account asset."""
        if data is None:
            if name is None:
                name = "test"
            match cloud_provider:
                case CloudProvidersType.AWS:
                    data = AWSProviderAccountData(
                        name=name,
                        configuration=AWSProviderAccountConfiguration(
                            region_name="test-region"
                        ),
                        credentials=AWSProviderAccountCredentials(
                            access_key_id="access-key",
                            secret_access_key="secret-key",
                        ),
                    )
                case CloudProvidersType.HETZNER:
                    data = HetznerProviderAccountData(
                        name=name,
                        configuration=HetznerProviderAccountConfiguration(
                            region_name="test-region",
                        ),
                        credentials=HetznerProviderAccountCredentials(
                            api_token="api-token",
                        ),
                    )
                case _:
                    data = DummyProviderAccountData(name=name)
        return self.create_asset(
            category=AssetCategory.CLOUD_PROVIDER_ACCOUNT,
            data=data,
            workspace=workspace,
            created_by=created_by,
            created_by_work_request=created_by_work_request,
        )

    def create_apt_authentication_asset(
        self,
        name: str | None = None,
        username: str | None = None,
        password: str | None = None,
        workspace: (
            Workspace | None | type[DEFAULT_WORKSPACE]
        ) = DEFAULT_WORKSPACE,
        created_by: User | None = None,
        created_by_work_request: WorkRequest | None = None,
    ) -> Asset:
        """Create a debusine:signing-key asset."""
        if workspace is DEFAULT_WORKSPACE:
            workspace = self.get_default_workspace()
        assert isinstance(workspace, Workspace) or workspace is None
        return self.create_asset(
            category=AssetCategory.APT_AUTHENTICATION,
            data=BasicAPTAuthenticationData(
                name=name or "test",
                credentials=BasicAPTCredentials(
                    username=username or "test-username",
                    password=password or "test-password",
                ),
            ),
            workspace=workspace,
            created_by=created_by,
            created_by_work_request=created_by_work_request,
        )

    @context.disable_permission_checks()
    def create_asset_usage(
        self,
        resource: Asset,
        workspace: Workspace | None = None,
    ) -> AssetUsage:
        """Create an AssetUsage model."""
        if workspace is None:
            workspace = self.get_default_workspace()
        return AssetUsage.objects.create(asset=resource, workspace=workspace)

    @context.disable_permission_checks()
    def create_debian_environments_collection(
        self, name: str = "debian", *, workspace: Workspace | None = None
    ) -> Collection:
        """Create a debian:environments collection."""
        if workspace is None:
            workspace = self.get_default_workspace()
        collection, _ = Collection.objects.get_or_create(
            name=name,
            category=CollectionCategory.ENVIRONMENTS,
            workspace=workspace,
        )
        return collection

    @context.disable_permission_checks()
    def create_debian_environment(  # noqa: C901
        self,
        *,
        environment: Artifact | None = None,
        category: ArtifactCategory = ArtifactCategory.SYSTEM_TARBALL,
        codename: str = "bookworm",
        architecture: str = "amd64",
        backend: BackendType = BackendType.UNSHARE,
        variant: str | None = None,
        collection: Collection | None = None,
        user: User | None = None,
        variables: dict[str, Any] | None = None,
        workspace: Workspace | None = None,
        vendor: str = "Debian",
        mirror: str = "https://deb.debian.org",
        components: list[str] | None = None,
        pkglist: dict[str, str] | None = None,
        with_init: bool = True,
        with_dev: bool = True,
        create_files: bool = False,
    ) -> CollectionItem:
        """Create a debian build environment."""
        if workspace is None:
            workspace = self.get_default_workspace()
        if collection is None:
            collection = self.create_debian_environments_collection(
                workspace=workspace
            )
        if user is None:
            user = self.get_default_user()
        if variant is None and variables is not None:
            variant = variables.get("variant")
        if environment is None:
            # Try looking up an existing environment
            lookup_string = (
                f"{collection.name}@{collection.category}"
                f"/match:codename={codename}"
                f":architecture={architecture}"
            )
            if category == ArtifactCategory.SYSTEM_TARBALL:
                lookup_string += ":format=tarball"
            else:
                lookup_string += ":format=image"
            if variant is not None:
                lookup_string += f":variant={variant}"
            try:
                lookup_result = lookup_single(
                    lookup_string,
                    workspace,
                    pc=PermissionContext.for_ui_user(user),
                )
                assert lookup_result.collection_item is not None
                return lookup_result.collection_item
            except KeyError:
                pass
        if components is None:
            components = ["main"]
        if environment is None:
            # Lookup failed: create it
            data = {
                "filename": "test",
                "vendor": vendor,
                "mirror": mirror,
                "pkglist": pkglist or {},
                "codename": codename,
                "components": components,
                "architecture": architecture,
                "variant": variant,
                "with_init": with_init,
                "with_dev": with_dev,
            }
            environment, _ = self.create_artifact(
                category=category,
                data=data,
                workspace=workspace,
                create_files=create_files,
            )
        if variables is None:
            variables = {"backend": backend, "variant": variant}

        try:
            return collection.child_items.active().get(
                artifact=environment, data__contains=variables
            )
        except CollectionItem.DoesNotExist:
            pass

        manager = collection.manager
        item = manager.add_artifact(environment, user=user, variables=variables)
        return item

    @context.disable_permission_checks()
    def create_sbuild_work_request(
        self,
        *,
        source: Artifact,
        architecture: str = "all",
        build_architecture: str | None = None,
        environment: Artifact,
        workflow: WorkRequest | None = None,
        workspace: Workspace | None = None,
    ) -> WorkRequest:
        """Create a sbuild work request."""
        task_data = SbuildData(
            input=SbuildInput(source_artifact=source.pk),
            build_architecture=build_architecture or architecture,
            environment=environment.pk,
            backend=BackendType.UNSHARE,
            build_components=[
                (
                    SbuildBuildComponent.ALL
                    if architecture == "all"
                    else SbuildBuildComponent.ANY
                )
            ],
        )
        if workflow:
            return workflow.create_child_worker(
                task_name="sbuild",
                task_data=task_data,
                workflow_data=WorkRequestWorkflowData(
                    display_name=f"Build {architecture}",
                    step=f"build-{architecture}",
                ),
            )
        else:
            assert task_data.build_architecture != "all", (
                "For an Sbuild task (not workflow): build_architecture must "
                "be a real architecture (e.g. amd64)"
            )
            return self.create_worker_task(
                task_name="sbuild",
                task_data=task_data,
                workspace=workspace or self.get_default_workspace(),
            )

    @context.disable_permission_checks()
    def simulate_package_build(
        self,
        source: Artifact,
        *,
        binaries: list[str] | None = None,
        architecture: str = "all",
        build_architecture: str | None = None,
        workflow: WorkRequest | None = None,
        environment: Artifact | None = None,
        worker: Worker | None = None,
    ) -> WorkRequest:
        """Generate database objects as if a package build happened."""
        workspace = source.workspace
        name = source.data["name"]
        version = source.data["version"]

        if environment is None:
            environment_item = self.create_debian_environment()
            assert environment_item.artifact is not None
            environment = environment_item.artifact

        build_arch = environment.data["architecture"]

        # Create sbuild work request
        work_request = self.create_sbuild_work_request(
            source=source,
            architecture=architecture,
            build_architecture=build_architecture,
            environment=environment,
            workflow=workflow,
            workspace=workspace,
        )
        self.advance_work_request(
            work_request, worker=worker, mark_running=True
        )

        # Add a build log
        buildlog_artifact = self.create_build_log_artifact(
            source=name,
            version=version,
            build_arch=build_arch,
            workspace=workspace,
            work_request=work_request,
        )
        self.create_artifact_relation(
            buildlog_artifact, source, ArtifactRelation.Relations.RELATES_TO
        )

        if binaries is None:
            binaries = [
                item["package"]
                for item in source.data["dsc_fields"]["Package-List"]
            ]

        # Add the binary upload
        upload_artifacts = self.create_upload_artifacts(
            src_name=name,
            version=version,
            source=False,
            binary=True,
            binaries=[(binary, architecture) for binary in binaries],
            workspace=workspace,
            work_request=work_request,
            created_by=work_request.created_by,
            create_files=True,
        )
        for binary in upload_artifacts.binaries:
            self.create_artifact_relation(
                binary,
                source,
                ArtifactRelation.Relations.BUILT_USING,
            )
            self.create_artifact_relation(
                buildlog_artifact,
                binary,
                ArtifactRelation.Relations.RELATES_TO,
            )
        self.create_artifact_relation(
            buildlog_artifact,
            upload_artifacts.upload,
            ArtifactRelation.Relations.RELATES_TO,
        )

        if workflow is not None:
            work_request.add_event_reaction(
                "on_success",
                ActionUpdateCollectionWithArtifacts(
                    collection="internal@collections",
                    name_template=f"build-{architecture}",
                    variables={
                        "binary_names": binaries,
                        "architecture": architecture,
                        "source_package_name": name,
                    },
                    artifact_filters={"category": ArtifactCategory.UPLOAD},
                ),
            )

        # Complete the task
        self.advance_work_request(
            work_request, result=WorkRequest.Results.SUCCESS
        )

        return work_request

    def build_scenario(
        self,
        scenario: "Scenario",
        /,
        scenario_name: str | None = None,
        set_current: bool = False,
        **kwargs: Any,
    ) -> None:
        """
        Build a test scenario.

        Keyword arguments are forwarded to the scenario constructor.

        :param scenario: scenario instance to build
        :param scenario_name: if set, register the result in self.scenarios
        :param set_current: set to True to set the current context from the
                            scenario
        :return: the scenario instance
        """
        with context.local():
            context.reset()
            scenario.build(self)
        if scenario_name is not None:
            self.scenarios[scenario_name] = scenario
        if set_current and scenario.needs_set_current:
            scenario.set_current()

    @context.disable_permission_checks()
    def create_pc_with_extra_groups(
        self,
        *groups: Group,
        user: User | None = None,
        workspace: Workspace | None = None,
    ) -> PermissionContext:
        """
        Create a PermissionContext with extra groups.

        This creates a supporting workflow template a work request.
        """
        workspace = workspace or self.get_default_workspace()
        wt_name = "create_pc_with_extra_groups" + str(
            self._seq_create_pc_with_extra_groups
        )
        self._seq_create_pc_with_extra_groups += 1
        wt = self.create_workflow_template(wt_name, "noop", workspace=workspace)
        wt.extra_groups.add(*groups)
        wr = self.create_workflow(
            wt,
            created_by=user or self.get_default_user(),
            workspace=workspace,
            mark_pending=False,
        )
        return PermissionContext.for_work_request(work_request=wr)
